# -*- coding: utf-8 -*-
#
#
# TheVirtualBrain-Framework Package. This package holds all Data Management, and 
# Web-UI helpful to run brain-simulations. To use it, you also need do download
# TheVirtualBrain-Scientific Package (for simulators). See content of the
# documentation-folder for more details. See also http://www.thevirtualbrain.org
#
# (c) 2012-2020, Baycrest Centre for Geriatric Care ("Baycrest") and others
#
# This program is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software Foundation,
# either version 3 of the License, or (at your option) any later version.
# This program is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
# PARTICULAR PURPOSE.  See the GNU General Public License for more details.
# You should have received a copy of the GNU General Public License along with this
# program.  If not, see <http://www.gnu.org/licenses/>.
#
#
#   CITATION:
# When using The Virtual Brain for scientific publications, please cite it as follows:
#
#   Paula Sanz Leon, Stuart A. Knock, M. Marmaduke Woodman, Lia Domide,
#   Jochen Mersmann, Anthony R. McIntosh, Viktor Jirsa (2013)
#       The Virtual Brain: a simulator of primate brain network dynamics.
#   Frontiers in Neuroinformatics (7:10. doi: 10.3389/fninf.2013.00010)
#
#

"""
Here we define entities related to workflows and portlets.

.. moduleauthor:: Lia Domide <lia.domide@codemart.ro>
.. moduleauthor:: Bogdan Neacsa <bogdan.neacsa@codemart.ro>
"""

import datetime
from sqlalchemy import Integer, String, Column, DateTime
from tvb.core.neotraits.db import Base


class Portlet(Base):
    """
    Store the Portlet entities. 
    One entity will hold:
    - portlet id;
    - path to the XML file that declares it;
    - a name given in the XML
    - an unique identifier (also from XML)
    - last date of introspection.
    """
    __tablename__ = 'PORTLETS'

    id = Column(Integer, primary_key=True)
    algorithm_identifier = Column(String)
    xml_path = Column(String)
    name = Column(String)
    last_introspection_check = Column(DateTime)


    def __init__(self, algorithm_identifier, xml_path, name="TVB Portlet"):
        self.algorithm_identifier = algorithm_identifier
        self.xml_path = xml_path
        self.name = name
        self.last_introspection_check = datetime.datetime.now()



# -*- coding: utf-8 -*-
#
#
# TheVirtualBrain-Framework Package. This package holds all Data Management, and 
# Web-UI helpful to run brain-simulations. To use it, you also need do download
# TheVirtualBrain-Scientific Package (for simulators). See content of the
# documentation-folder for more details. See also http://www.thevirtualbrain.org
#
# (c) 2012-2020, Baycrest Centre for Geriatric Care ("Baycrest") and others
#
# This program is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software Foundation,
# either version 3 of the License, or (at your option) any later version.
# This program is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
# PARTICULAR PURPOSE.  See the GNU General Public License for more details.
# You should have received a copy of the GNU General Public License along with this
# program.  If not, see <http://www.gnu.org/licenses/>.
#
#
#   CITATION:
# When using The Virtual Brain for scientific publications, please cite it as follows:
#
#   Paula Sanz Leon, Stuart A. Knock, M. Marmaduke Woodman, Lia Domide,
#   Jochen Mersmann, Anthony R. McIntosh, Viktor Jirsa (2013)
#       The Virtual Brain: a simulator of primate brain network dynamics.
#   Frontiers in Neuroinformatics (7:10. doi: 10.3389/fninf.2013.00010)
#
#

"""
.. moduleauthor:: Marmaduke Woodman <marmaduke.woodman@univ-amu.fr>
.. moduleauthor:: Lia Domide <lia.domide@codemart.ro>
"""

from tvb.basic.logger.builder import get_logger
from tvb.core.adapters.exceptions import LaunchException
from tvb.core.adapters.abcuploader import ABCUploader, ABCUploaderForm
from tvb.core.entities.filters.chain import FilterChain
from tvb.adapters.datatypes.db.projections import ProjectionMatrixIndex
from tvb.adapters.datatypes.db.sensors import SensorsIndex
from tvb.core.neotraits.forms import TraitUploadField, StrField, TraitDataTypeSelectField
from tvb.core.neocom import h5
from tvb.core.neotraits.uploader_view_model import UploaderViewModel
from tvb.core.neotraits.view_model import Str, DataTypeGidAttr
from tvb.datatypes.sensors import SensorsEEG, SensorsMEG, Sensors
from tvb.datatypes.projections import *
from tvb.datatypes.surfaces import CorticalSurface, Surface

DEFAULT_DATASET_NAME = "ProjectionMatrix"


def determine_projection_type(sensors_idx):
    # type: (SensorsIndex) -> str
    if sensors_idx.sensors_type == SensorsEEG.sensors_type.default:
        projection_matrix_type = ProjectionSurfaceEEG.projection_type.default
    elif sensors_idx.sensors_type == SensorsMEG.sensors_type.default:
        projection_matrix_type = ProjectionSurfaceMEG.projection_type.default
    else:
        projection_matrix_type = ProjectionSurfaceSEEG.projection_type.default

    return projection_matrix_type


class ProjectionMatrixImporterModel(UploaderViewModel):
    projection_file = Str(
        label='Projection matrix file (.mat or .npy format)',
        doc='Expected a file containing projection matrix (one vector of length '
            'number of surface vertices nd values in the sensors range).'
    )

    dataset_name = Attr(
        field_type=str,
        required=False,
        default=DEFAULT_DATASET_NAME,
        label='Matlab dataset name',
        doc='Name of the MATLAB dataset where data is stored. Required only for .mat files'
    )

    surface = DataTypeGidAttr(
        linked_datatype=Surface,
        label='Brain Cortical Surface',
        doc='The Brain Surface used by the uploaded projection matrix.'
    )

    sensors = DataTypeGidAttr(
        linked_datatype=Sensors,
        label='Sensors',
        doc='The Sensors used in for current projection.'
    )


class ProjectionMatrixImporterForm(ABCUploaderForm):

    def __init__(self, prefix='', project_id=None):
        super(ProjectionMatrixImporterForm, self).__init__(prefix, project_id)
        self.projection_file = TraitUploadField(ProjectionMatrixImporterModel.projection_file, '.mat, .npy', self,
                                                name='projection_file')
        self.dataset_name = StrField(ProjectionMatrixImporterModel.dataset_name, self, name='dataset_name')
        surface_conditions = FilterChain(fields=[FilterChain.datatype + '.surface_type'], operations=['=='],
                                         values=['Cortical Surface'])
        self.surface = TraitDataTypeSelectField(ProjectionMatrixImporterModel.surface, self, name='surface',
                                                conditions=surface_conditions)
        self.sensors = TraitDataTypeSelectField(ProjectionMatrixImporterModel.sensors, self, name='sensors')

    @staticmethod
    def get_view_model():
        return ProjectionMatrixImporterModel


class ProjectionMatrixSurfaceEEGImporter(ABCUploader):
    """
    Upload ProjectionMatrix Cortical Surface -> EEG/MEG/SEEG Sensors from a MAT or NPY file.
    """

    _ui_name = "Gain Matrix for Sensors"
    _ui_description = "Upload a Projection Matrix between a Brain Cortical Surface and EEG/MEG Sensors."
    logger = get_logger(__name__)

    def get_form_class(self):
        return ProjectionMatrixImporterForm

    def get_output(self):
        return [ProjectionMatrixIndex]

    def launch(self, view_model):
        # type: (ProjectionMatrixImporterModel) -> [ProjectionMatrixIndex]
        """
        Creates ProjectionMatrix entity from uploaded data.

        :raises LaunchException: when
                    * no projection_file or sensors are specified
                    * the dataset is invalid
                    * number of sensors is different from the one in dataset
        """
        if view_model.projection_file is None:
            raise LaunchException("Please select MATLAB file which contains data to import")

        if view_model.sensors is None:
            raise LaunchException("No sensors selected. Please initiate upload again and select one.")

        if view_model.surface is None:
            raise LaunchException("No source selected. Please initiate upload again and select a source.")

        surface_index = self.load_entity_by_gid(view_model.surface.hex)
        expected_surface_shape = surface_index.number_of_vertices

        sensors_index = self.load_entity_by_gid(view_model.sensors.hex)
        expected_sensors_shape = sensors_index.number_of_sensors

        self.logger.debug("Reading projection matrix from uploaded file...")
        if view_model.projection_file.endswith(".mat"):
            projection_data = self.read_matlab_data(view_model.projection_file, view_model.dataset_name)
        else:
            projection_data = self.read_list_data(view_model.projection_file)

        if projection_data is None or len(projection_data) == 0:
            raise LaunchException("Invalid (empty) dataset...")

        if projection_data.shape[0] != expected_sensors_shape:
            raise LaunchException("Invalid Projection Matrix shape[0]: %d Expected: %d" % (projection_data.shape[0],
                                                                                           expected_sensors_shape))

        if projection_data.shape[1] != expected_surface_shape:
            raise LaunchException("Invalid Projection Matrix shape[1]: %d Expected: %d" % (projection_data.shape[1],
                                                                                           expected_surface_shape))

        projection_matrix_type = determine_projection_type(sensors_index)
        surface_ht = h5.load_from_index(surface_index, CorticalSurface)
        sensors_ht = h5.load_from_index(sensors_index)
        projection_matrix = ProjectionMatrix(sources=surface_ht, sensors=sensors_ht,
                                             projection_type=projection_matrix_type,
                                             projection_data=projection_data)
        return h5.store_complete(projection_matrix, self.storage_path)



# -*- coding: utf-8 -*-
#
#
# TheVirtualBrain-Framework Package. This package holds all Data Management, and
# Web-UI helpful to run brain-simulations. To use it, you also need do download
# TheVirtualBrain-Scientific Package (for simulators). See content of the
# documentation-folder for more details. See also http://www.thevirtualbrain.org
#
# (c) 2012-2020, Baycrest Centre for Geriatric Care ("Baycrest") and others
#
# This program is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software Foundation,
# either version 3 of the License, or (at your option) any later version.
# This program is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
# PARTICULAR PURPOSE.  See the GNU General Public License for more details.
# You should have received a copy of the GNU General Public License along with this
# program.  If not, see <http://www.gnu.org/licenses/>.
#
#
#   CITATION:
# When using The Virtual Brain for scientific publications, please cite it as follows:
#
#   Paula Sanz Leon, Stuart A. Knock, M. Marmaduke Woodman, Lia Domide,
#   Jochen Mersmann, Anthony R. McIntosh, Viktor Jirsa (2013)
#       The Virtual Brain: a simulator of primate brain network dynamics.
#   Frontiers in Neuroinformatics (7:10. doi: 10.3389/fninf.2013.00010)
#
#
import json
import numpy
from tvb.simulator.models import *
from tvb.core.entities.file.simulator.configurations_h5 import SimulatorConfigurationH5
from tvb.core.neotraits.h5 import DataSet, Json, JsonFinal


class StateVariablesEncoder(json.JSONEncoder):
    def default(self, o):
        if isinstance(o, numpy.ndarray):
            o = o.tolist()
        return o


class StateVariablesDecoder(json.JSONDecoder):
    def __init__(self):
        json.JSONDecoder.__init__(self, object_hook=self.dict_array)

    def dict_array(self, dictionary):
        dict_array = {}
        for k, v in dictionary.items():
            dict_array.update({k: numpy.array(v)})
        return dict_array


class EpileptorH5(SimulatorConfigurationH5):

    def __init__(self, path):
        super(EpileptorH5, self).__init__(path)

        self.a = DataSet(Epileptor.a, self)
        self.b = DataSet(Epileptor.b, self)
        self.c = DataSet(Epileptor.c, self)
        self.d = DataSet(Epileptor.d, self)
        self.r = DataSet(Epileptor.r, self)
        self.s = DataSet(Epileptor.s, self)
        self.x0 = DataSet(Epileptor.x0, self)
        self.Iext = DataSet(Epileptor.Iext, self)
        self.slope = DataSet(Epileptor.slope, self)
        self.Iext2 = DataSet(Epileptor.Iext2, self)
        self.tau = DataSet(Epileptor.tau, self)
        self.aa = DataSet(Epileptor.aa, self)
        self.bb = DataSet(Epileptor.bb, self)
        self.Kvf = DataSet(Epileptor.Kvf, self)
        self.Kf = DataSet(Epileptor.Kf, self)
        self.Ks = DataSet(Epileptor.Ks, self)
        self.tt = DataSet(Epileptor.tt, self)
        self.modification = DataSet(Epileptor.modification, self)
        self.state_variable_range = JsonFinal(Epileptor.state_variable_range, self, json_encoder=StateVariablesEncoder,
                                              json_decoder=StateVariablesDecoder)
        self.variables_of_interest = Json(Epileptor.variables_of_interest, self)


class Epileptor2DH5(SimulatorConfigurationH5):

    def __init__(self, path):
        super(Epileptor2DH5, self).__init__(path)

        self.a = DataSet(Epileptor2D.a, self)
        self.b = DataSet(Epileptor2D.b, self)
        self.c = DataSet(Epileptor2D.c, self)
        self.d = DataSet(Epileptor2D.d, self)
        self.r = DataSet(Epileptor2D.r, self)
        self.x0 = DataSet(Epileptor2D.x0, self)
        self.Iext = DataSet(Epileptor2D.Iext, self)
        self.slope = DataSet(Epileptor2D.slope, self)
        self.Kvf = DataSet(Epileptor2D.Kvf, self)
        self.Ks = DataSet(Epileptor2D.Ks, self)
        self.tt = DataSet(Epileptor2D.tt, self)
        self.modification = DataSet(Epileptor2D.modification, self)
        self.state_variable_range = JsonFinal(Epileptor2D.state_variable_range, self,
                                              json_encoder=StateVariablesEncoder, json_decoder=StateVariablesDecoder)
        self.variables_of_interest = Json(Epileptor2D.variables_of_interest, self)


class EpileptorCodim3H5(SimulatorConfigurationH5):

    def __init__(self, path):
        super(EpileptorCodim3H5, self).__init__(path)
        self.mu1_start = DataSet(EpileptorCodim3.mu1_start, self)
        self.mu2_start = DataSet(EpileptorCodim3.mu2_start, self)
        self.nu_start = DataSet(EpileptorCodim3.nu_start, self)
        self.mu1_stop = DataSet(EpileptorCodim3.mu1_stop, self)
        self.mu2_stop = DataSet(EpileptorCodim3.mu2_stop, self)
        self.nu_stop = DataSet(EpileptorCodim3.nu_stop, self)
        self.b = DataSet(EpileptorCodim3.b, self)
        self.R = DataSet(EpileptorCodim3.R, self)
        self.c = DataSet(EpileptorCodim3.c, self)
        self.dstar = DataSet(EpileptorCodim3.dstar, self)
        self.Ks = DataSet(EpileptorCodim3.Ks, self)
        self.N = DataSet(EpileptorCodim3.N, self)
        self.modification = DataSet(EpileptorCodim3.modification, self)
        self.state_variable_range = JsonFinal(EpileptorCodim3.state_variable_range, self,
                                              json_encoder=StateVariablesEncoder, json_decoder=StateVariablesDecoder)
        self.variables_of_interest = Json(EpileptorCodim3.variables_of_interest, self)


class EpileptorCodim3SlowModH5(SimulatorConfigurationH5):

    def __init__(self, path):
        super(EpileptorCodim3SlowModH5, self).__init__(path)
        self.mu1_Ain = DataSet(EpileptorCodim3SlowMod.mu1_Ain, self)
        self.mu2_Ain = DataSet(EpileptorCodim3SlowMod.mu2_Ain, self)
        self.nu_Ain = DataSet(EpileptorCodim3SlowMod.nu_Ain, self)
        self.mu1_Bin = DataSet(EpileptorCodim3SlowMod.mu1_Bin, self)
        self.mu2_Bin = DataSet(EpileptorCodim3SlowMod.mu2_Bin, self)
        self.nu_Bin = DataSet(EpileptorCodim3SlowMod.nu_Bin, self)
        self.mu1_Aend = DataSet(EpileptorCodim3SlowMod.mu1_Aend, self)
        self.mu2_Aend = DataSet(EpileptorCodim3SlowMod.mu2_Aend, self)
        self.nu_Aend = DataSet(EpileptorCodim3SlowMod.nu_Aend, self)
        self.mu1_Bend = DataSet(EpileptorCodim3SlowMod.mu1_Bend, self)
        self.mu2_Bend = DataSet(EpileptorCodim3SlowMod.mu2_Bend, self)
        self.nu_Bend = DataSet(EpileptorCodim3SlowMod.nu_Bend, self)
        self.b = DataSet(EpileptorCodim3SlowMod.b, self)
        self.R = DataSet(EpileptorCodim3SlowMod.R, self)
        self.c = DataSet(EpileptorCodim3SlowMod.c, self)
        self.cA = DataSet(EpileptorCodim3SlowMod.cA, self)
        self.cB = DataSet(EpileptorCodim3SlowMod.cB, self)
        self.dstar = DataSet(EpileptorCodim3SlowMod.dstar, self)
        self.Ks = DataSet(EpileptorCodim3SlowMod.Ks, self)
        self.N = DataSet(EpileptorCodim3SlowMod.N, self)
        self.modification = DataSet(EpileptorCodim3SlowMod.modification, self)
        self.state_variable_range = JsonFinal(EpileptorCodim3SlowMod.state_variable_range, self,
                                              json_encoder=StateVariablesEncoder, json_decoder=StateVariablesDecoder)
        self.variables_of_interest = Json(EpileptorCodim3SlowMod.variables_of_interest, self)


class HopfieldH5(SimulatorConfigurationH5):

    def __init__(self, path):
        super(HopfieldH5, self).__init__(path)
        self.taux = DataSet(Hopfield.taux, self)
        self.tauT = DataSet(Hopfield.tauT, self)
        self.dynamic = DataSet(Hopfield.dynamic, self)
        self.state_variable_range = JsonFinal(Hopfield.state_variable_range, self, json_encoder=StateVariablesEncoder,
                                              json_decoder=StateVariablesDecoder)
        self.variables_of_interest = Json(Hopfield.variables_of_interest, self)


class JansenRitH5(SimulatorConfigurationH5):

    def __init__(self, path):
        super(JansenRitH5, self).__init__(path)
        self.A = DataSet(JansenRit.A, self)
        self.B = DataSet(JansenRit.B, self)
        self.a = DataSet(JansenRit.a, self)
        self.b = DataSet(JansenRit.b, self)
        self.v0 = DataSet(JansenRit.v0, self)
        self.nu_max = DataSet(JansenRit.nu_max, self)
        self.r = DataSet(JansenRit.r, self)
        self.J = DataSet(JansenRit.J, self)
        self.a_1 = DataSet(JansenRit.a_1, self)
        self.a_2 = DataSet(JansenRit.a_2, self)
        self.a_3 = DataSet(JansenRit.a_3, self)
        self.a_4 = DataSet(JansenRit.a_4, self)
        self.p_min = DataSet(JansenRit.p_min, self)
        self.p_max = DataSet(JansenRit.p_max, self)
        self.mu = DataSet(JansenRit.mu, self)
        self.state_variable_range = JsonFinal(JansenRit.state_variable_range, self, json_encoder=StateVariablesEncoder,
                                              json_decoder=StateVariablesDecoder)
        self.variables_of_interest = Json(JansenRit.variables_of_interest, self)


class ZetterbergJansenH5(SimulatorConfigurationH5):

    def __init__(self, path):
        super(ZetterbergJansenH5, self).__init__(path)
        self.He = DataSet(ZetterbergJansen.He, self)
        self.Hi = DataSet(ZetterbergJansen.Hi, self)
        self.ke = DataSet(ZetterbergJansen.ke, self)
        self.ki = DataSet(ZetterbergJansen.ki, self)
        self.e0 = DataSet(ZetterbergJansen.e0, self)
        self.rho_2 = DataSet(ZetterbergJansen.rho_2, self)
        self.rho_1 = DataSet(ZetterbergJansen.rho_1, self)
        self.gamma_1 = DataSet(ZetterbergJansen.gamma_1, self)
        self.gamma_2 = DataSet(ZetterbergJansen.gamma_2, self)
        self.gamma_3 = DataSet(ZetterbergJansen.gamma_3, self)
        self.gamma_4 = DataSet(ZetterbergJansen.gamma_4, self)
        self.gamma_5 = DataSet(ZetterbergJansen.gamma_5, self)
        self.gamma_1T = DataSet(ZetterbergJansen.gamma_1T, self)
        self.gamma_2T = DataSet(ZetterbergJansen.gamma_2T, self)
        self.gamma_3T = DataSet(ZetterbergJansen.gamma_3T, self)
        self.P = DataSet(ZetterbergJansen.P, self)
        self.U = DataSet(ZetterbergJansen.U, self)
        self.Q = DataSet(ZetterbergJansen.Q, self)
        self.state_variable_range = JsonFinal(ZetterbergJansen.state_variable_range, self,
                                              json_encoder=StateVariablesEncoder, json_decoder=StateVariablesDecoder)
        self.variables_of_interest = Json(ZetterbergJansen.variables_of_interest, self)


class EpileptorRestingStateH5(SimulatorConfigurationH5):

    def __init__(self, path):
        super(EpileptorRestingStateH5, self).__init__(path)
        self.a = DataSet(EpileptorRestingState.a, self)
        self.b = DataSet(EpileptorRestingState.b, self)
        self.c = DataSet(EpileptorRestingState.c, self)
        self.d = DataSet(EpileptorRestingState.d, self)
        self.r = DataSet(EpileptorRestingState.r, self)
        self.s = DataSet(EpileptorRestingState.s, self)
        self.x0 = DataSet(EpileptorRestingState.x0, self)
        self.Iext = DataSet(EpileptorRestingState.Iext, self)
        self.slope = DataSet(EpileptorRestingState.slope, self)
        self.Iext2 = DataSet(EpileptorRestingState.Iext2, self)
        self.tau = DataSet(EpileptorRestingState.tau, self)
        self.aa = DataSet(EpileptorRestingState.aa, self)
        self.bb = DataSet(EpileptorRestingState.bb, self)
        self.Kvf = DataSet(EpileptorRestingState.Kvf, self)
        self.Kf = DataSet(EpileptorRestingState.Kf, self)
        self.Ks = DataSet(EpileptorRestingState.Ks, self)
        self.tt = DataSet(EpileptorRestingState.tt, self)
        self.tau_rs = DataSet(EpileptorRestingState.tau_rs, self)
        self.I_rs = DataSet(EpileptorRestingState.I_rs, self)
        self.a_rs = DataSet(EpileptorRestingState.a_rs, self)
        self.b_rs = DataSet(EpileptorRestingState.b_rs, self)
        self.d_rs = DataSet(EpileptorRestingState.d_rs, self)
        self.e_rs = DataSet(EpileptorRestingState.e_rs, self)
        self.f_rs = DataSet(EpileptorRestingState.f_rs, self)
        self.alpha_rs = DataSet(EpileptorRestingState.alpha_rs, self)
        self.beta_rs = DataSet(EpileptorRestingState.beta_rs, self)
        self.gamma_rs = DataSet(EpileptorRestingState.gamma_rs, self)
        self.K_rs = DataSet(EpileptorRestingState.K_rs, self)
        self.p = DataSet(EpileptorRestingState.p, self)
        self.state_variable_range = JsonFinal(EpileptorRestingState.state_variable_range, self,
                                              json_encoder=StateVariablesEncoder, json_decoder=StateVariablesDecoder)
        self.variables_of_interest = Json(EpileptorRestingState.variables_of_interest, self)


class LarterBreakspearH5(SimulatorConfigurationH5):

    def __init__(self, path):
        super(LarterBreakspearH5, self).__init__(path)
        self.gCa = DataSet(LarterBreakspear.gCa, self)
        self.gK = DataSet(LarterBreakspear.gK, self)
        self.gL = DataSet(LarterBreakspear.gL, self)
        self.phi = DataSet(LarterBreakspear.phi, self)
        self.gNa = DataSet(LarterBreakspear.gNa, self)
        self.TK = DataSet(LarterBreakspear.TK, self)
        self.TCa = DataSet(LarterBreakspear.TCa, self)
        self.TNa = DataSet(LarterBreakspear.TNa, self)
        self.VCa = DataSet(LarterBreakspear.VCa, self)
        self.VK = DataSet(LarterBreakspear.VK, self)
        self.VL = DataSet(LarterBreakspear.VL, self)
        self.VNa = DataSet(LarterBreakspear.VNa, self)
        self.d_K = DataSet(LarterBreakspear.d_K, self)
        self.tau_K = DataSet(LarterBreakspear.tau_K, self)
        self.d_Na = DataSet(LarterBreakspear.d_Na, self)
        self.d_Ca = DataSet(LarterBreakspear.d_Ca, self)
        self.aei = DataSet(LarterBreakspear.aei, self)
        self.aie = DataSet(LarterBreakspear.aie, self)
        self.b = DataSet(LarterBreakspear.b, self)
        self.C = DataSet(LarterBreakspear.C, self)
        self.ane = DataSet(LarterBreakspear.ane, self)
        self.ani = DataSet(LarterBreakspear.ani, self)
        self.aee = DataSet(LarterBreakspear.aee, self)
        self.Iext = DataSet(LarterBreakspear.Iext, self)
        self.rNMDA = DataSet(LarterBreakspear.rNMDA, self)
        self.VT = DataSet(LarterBreakspear.VT, self)
        self.d_V = DataSet(LarterBreakspear.d_V, self)
        self.ZT = DataSet(LarterBreakspear.ZT, self)
        self.d_Z = DataSet(LarterBreakspear.d_Z, self)
        self.QV_max = DataSet(LarterBreakspear.QV_max, self)
        self.QZ_max = DataSet(LarterBreakspear.QZ_max, self)
        self.t_scale = DataSet(LarterBreakspear.t_scale, self)
        self.variables_of_interest = Json(LarterBreakspear.variables_of_interest, self)
        self.state_variable_range = JsonFinal(LarterBreakspear.state_variable_range, self,
                                              json_encoder=StateVariablesEncoder, json_decoder=StateVariablesDecoder)


class LinearH5(SimulatorConfigurationH5):

    def __init__(self, path):
        super(LinearH5, self).__init__(path)
        self.gamma = DataSet(Linear.gamma, self)
        self.state_variable_range = JsonFinal(Linear.state_variable_range, self, json_encoder=StateVariablesEncoder,
                                              json_decoder=StateVariablesDecoder)
        self.variables_of_interest = Json(Linear.variables_of_interest, self)


class Generic2dOscillatorH5(SimulatorConfigurationH5):

    def __init__(self, path):
        super(Generic2dOscillatorH5, self).__init__(path)
        self.tau = DataSet(Generic2dOscillator.tau, self)
        self.I = DataSet(Generic2dOscillator.I, self)
        self.a = DataSet(Generic2dOscillator.a, self)
        self.b = DataSet(Generic2dOscillator.b, self)
        self.c = DataSet(Generic2dOscillator.c, self)
        self.d = DataSet(Generic2dOscillator.d, self)
        self.e = DataSet(Generic2dOscillator.e, self)
        self.f = DataSet(Generic2dOscillator.f, self)
        self.g = DataSet(Generic2dOscillator.g, self)
        self.alpha = DataSet(Generic2dOscillator.alpha, self)
        self.beta = DataSet(Generic2dOscillator.beta, self)
        self.gamma = DataSet(Generic2dOscillator.gamma, self)
        self.state_variable_range = JsonFinal(Generic2dOscillator.state_variable_range, self,
                                              json_encoder=StateVariablesEncoder, json_decoder=StateVariablesDecoder)
        self.variables_of_interest = Json(Generic2dOscillator.variables_of_interest, self)


class KuramotoH5(SimulatorConfigurationH5):

    def __init__(self, path):
        super(KuramotoH5, self).__init__(path)
        self.omega = DataSet(Kuramoto.omega, self)
        self.state_variable_range = JsonFinal(Kuramoto.state_variable_range, self, json_encoder=StateVariablesEncoder,
                                              json_decoder=StateVariablesDecoder)
        self.variables_of_interest = Json(Kuramoto.variables_of_interest, self)


class SupHopfH5(SimulatorConfigurationH5):

    def __init__(self, path):
        super(SupHopfH5, self).__init__(path)
        self.a = DataSet(SupHopf.a, self)
        self.omega = DataSet(SupHopf.omega, self)
        self.state_variable_range = JsonFinal(SupHopf.state_variable_range, self, json_encoder=StateVariablesEncoder,
                                              json_decoder=StateVariablesDecoder)
        self.variables_of_interest = Json(SupHopf.variables_of_interest, self)


class ReducedSetFitzHughNagumoH5(SimulatorConfigurationH5):

    def __init__(self, path):
        super(ReducedSetFitzHughNagumoH5, self).__init__(path)
        self.tau = DataSet(ReducedSetFitzHughNagumo.tau, self)
        self.a = DataSet(ReducedSetFitzHughNagumo.a, self)
        self.b = DataSet(ReducedSetFitzHughNagumo.b, self)
        self.K11 = DataSet(ReducedSetFitzHughNagumo.K11, self)
        self.K12 = DataSet(ReducedSetFitzHughNagumo.K12, self)
        self.K21 = DataSet(ReducedSetFitzHughNagumo.K21, self)
        self.sigma = DataSet(ReducedSetFitzHughNagumo.sigma, self)
        self.mu = DataSet(ReducedSetFitzHughNagumo.mu, self)
        self.state_variable_range = JsonFinal(ReducedSetFitzHughNagumo.state_variable_range, self,
                                              json_encoder=StateVariablesEncoder, json_decoder=StateVariablesDecoder)
        self.variables_of_interest = Json(ReducedSetFitzHughNagumo.variables_of_interest, self)


class ReducedSetHindmarshRoseH5(SimulatorConfigurationH5):

    def __init__(self, path):
        super(ReducedSetHindmarshRoseH5, self).__init__(path)
        self.r = DataSet(ReducedSetHindmarshRose.r, self)
        self.a = DataSet(ReducedSetHindmarshRose.a, self)
        self.b = DataSet(ReducedSetHindmarshRose.b, self)
        self.c = DataSet(ReducedSetHindmarshRose.c, self)
        self.d = DataSet(ReducedSetHindmarshRose.d, self)
        self.s = DataSet(ReducedSetHindmarshRose.s, self)
        self.xo = DataSet(ReducedSetHindmarshRose.xo, self)
        self.K11 = DataSet(ReducedSetHindmarshRose.K11, self)
        self.K12 = DataSet(ReducedSetHindmarshRose.K12, self)
        self.K21 = DataSet(ReducedSetHindmarshRose.K21, self)
        self.sigma = DataSet(ReducedSetHindmarshRose.sigma, self)
        self.mu = DataSet(ReducedSetHindmarshRose.mu, self)
        self.state_variable_range = JsonFinal(ReducedSetHindmarshRose.state_variable_range, self,
                                              json_encoder=StateVariablesEncoder, json_decoder=StateVariablesDecoder)
        self.variables_of_interest = Json(ReducedSetHindmarshRose.variables_of_interest, self)


class WilsonCowanH5(SimulatorConfigurationH5):

    def __init__(self, path):
        super(WilsonCowanH5, self).__init__(path)
        self.c_ee = DataSet(WilsonCowan.c_ee, self)
        self.c_ie = DataSet(WilsonCowan.c_ie, self)
        self.c_ei = DataSet(WilsonCowan.c_ei, self)
        self.c_ii = DataSet(WilsonCowan.c_ii, self)
        self.tau_e = DataSet(WilsonCowan.tau_e, self)
        self.tau_i = DataSet(WilsonCowan.tau_i, self)
        self.a_e = DataSet(WilsonCowan.a_e, self)
        self.b_e = DataSet(WilsonCowan.b_e, self)
        self.c_e = DataSet(WilsonCowan.c_e, self)
        self.theta_e = DataSet(WilsonCowan.theta_e, self)
        self.a_i = DataSet(WilsonCowan.a_i, self)
        self.b_i = DataSet(WilsonCowan.b_i, self)
        self.theta_i = DataSet(WilsonCowan.theta_i, self)
        self.c_i = DataSet(WilsonCowan.c_i, self)
        self.r_e = DataSet(WilsonCowan.r_e, self)
        self.r_i = DataSet(WilsonCowan.r_i, self)
        self.k_e = DataSet(WilsonCowan.k_e, self)
        self.k_i = DataSet(WilsonCowan.k_i, self)
        self.P = DataSet(WilsonCowan.P, self)
        self.Q = DataSet(WilsonCowan.Q, self)
        self.alpha_e = DataSet(WilsonCowan.alpha_e, self)
        self.alpha_i = DataSet(WilsonCowan.alpha_i, self)
        self.state_variable_range = JsonFinal(WilsonCowan.state_variable_range, self,
                                              json_encoder=StateVariablesEncoder, json_decoder=StateVariablesDecoder)
        self.variables_of_interest = Json(WilsonCowan.variables_of_interest, self)


class ReducedWongWangH5(SimulatorConfigurationH5):

    def __init__(self, path):
        super(ReducedWongWangH5, self).__init__(path)
        self.a = DataSet(ReducedWongWang.a, self)
        self.b = DataSet(ReducedWongWang.b, self)
        self.d = DataSet(ReducedWongWang.d, self)
        self.gamma = DataSet(ReducedWongWang.gamma, self)
        self.tau_s = DataSet(ReducedWongWang.tau_s, self)
        self.w = DataSet(ReducedWongWang.w, self)
        self.J_N = DataSet(ReducedWongWang.J_N, self)
        self.I_o = DataSet(ReducedWongWang.I_o, self)
        self.sigma_noise = DataSet(ReducedWongWang.sigma_noise, self)
        self.state_variable_range = JsonFinal(ReducedWongWang.state_variable_range, self,
                                              json_encoder=StateVariablesEncoder, json_decoder=StateVariablesDecoder)
        self.variables_of_interest = Json(ReducedWongWang.variables_of_interest, self)


class ReducedWongWangExcInhH5(SimulatorConfigurationH5):

    def __init__(self, path):
        super(ReducedWongWangExcInhH5, self).__init__(path)
        self.a_e = DataSet(ReducedWongWangExcInh.a_e, self)
        self.b_e = DataSet(ReducedWongWangExcInh.b_e, self)
        self.d_e = DataSet(ReducedWongWangExcInh.d_e, self)
        self.gamma_e = DataSet(ReducedWongWangExcInh.gamma_e, self)
        self.tau_e = DataSet(ReducedWongWangExcInh.tau_e, self)
        self.w_p = DataSet(ReducedWongWangExcInh.w_p, self)
        self.J_N = DataSet(ReducedWongWangExcInh.J_N, self)
        self.W_e = DataSet(ReducedWongWangExcInh.W_e, self)
        self.a_i = DataSet(ReducedWongWangExcInh.a_i, self)
        self.b_i = DataSet(ReducedWongWangExcInh.b_i, self)
        self.d_i = DataSet(ReducedWongWangExcInh.d_i, self)
        self.gamma_i = DataSet(ReducedWongWangExcInh.gamma_i, self)
        self.tau_i = DataSet(ReducedWongWangExcInh.tau_i, self)
        self.J_i = DataSet(ReducedWongWangExcInh.J_i, self)
        self.W_i = DataSet(ReducedWongWangExcInh.W_i, self)
        self.I_o = DataSet(ReducedWongWangExcInh.I_o, self)
        self.G = DataSet(ReducedWongWangExcInh.G, self)
        self.lamda = DataSet(ReducedWongWangExcInh.lamda, self)
        self.state_variable_range = JsonFinal(ReducedWongWangExcInh.state_variable_range, self,
                                              json_encoder=StateVariablesEncoder, json_decoder=StateVariablesDecoder)
        self.variables_of_interest = Json(ReducedWongWangExcInh.variables_of_interest, self)


class ZerlautFirstOrderH5(SimulatorConfigurationH5):

    def __init__(self, path):
        super(ZerlautFirstOrderH5, self).__init__(path)
        self.g_L = DataSet(ZerlautFirstOrder.g_L, self)
        self.E_L_e = DataSet(ZerlautFirstOrder.E_L_e, self)
        self.E_L_i = DataSet(ZerlautFirstOrder.E_L_i, self)
        self.C_m = DataSet(ZerlautFirstOrder.C_m, self)
        self.b = DataSet(ZerlautFirstOrder.b, self)
        self.tau_w = DataSet(ZerlautFirstOrder.tau_w, self)
        self.E_e = DataSet(ZerlautFirstOrder.E_e, self)
        self.E_i = DataSet(ZerlautFirstOrder.E_i, self)
        self.Q_e = DataSet(ZerlautFirstOrder.Q_e, self)
        self.Q_i = DataSet(ZerlautFirstOrder.Q_i, self)
        self.tau_e = DataSet(ZerlautFirstOrder.tau_e, self)
        self.tau_i = DataSet(ZerlautFirstOrder.tau_i, self)
        self.N_tot = DataSet(ZerlautFirstOrder.N_tot, self)
        self.p_connect = DataSet(ZerlautFirstOrder.p_connect, self)
        self.g = DataSet(ZerlautFirstOrder.g, self)
        self.T = DataSet(ZerlautFirstOrder.T, self)
        self.P_e = DataSet(ZerlautFirstOrder.P_e, self)
        self.P_i = DataSet(ZerlautFirstOrder.P_i, self)
        self.external_input = DataSet(ZerlautFirstOrder.external_input, self)
        self.state_variable_range = JsonFinal(ZerlautFirstOrder.state_variable_range, self,
                                              json_encoder=StateVariablesEncoder, json_decoder=StateVariablesDecoder)
        self.variables_of_interest = Json(ZerlautFirstOrder.variables_of_interest, self)


class ZerlautSecondOrderH5(ZerlautFirstOrderH5):

    def __init__(self, path):
        super(ZerlautSecondOrderH5, self).__init__(path)
        self.state_variable_range = JsonFinal(ZerlautSecondOrder.state_variable_range, self,
                                              json_encoder=StateVariablesEncoder, json_decoder=StateVariablesDecoder)
        self.variables_of_interest = Json(ZerlautSecondOrder.variables_of_interest, self)



# -*- coding: utf-8 -*-
#
#
# TheVirtualBrain-Framework Package. This package holds all Data Management, and
# Web-UI helpful to run brain-simulations. To use it, you also need do download
# TheVirtualBrain-Scientific Package (for simulators). See content of the
# documentation-folder for more details. See also http://www.thevirtualbrain.org
#
# (c) 2012-2020, Baycrest Centre for Geriatric Care ("Baycrest") and others
#
# This program is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software Foundation,
# either version 3 of the License, or (at your option) any later version.
# This program is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
# PARTICULAR PURPOSE.  See the GNU General Public License for more details.
# You should have received a copy of the GNU General Public License along with this
# program.  If not, see <http://www.gnu.org/licenses/>.
#
#
#   CITATION:
# When using The Virtual Brain for scientific publications, please cite it as follows:
#
#   Paula Sanz Leon, Stuart A. Knock, M. Marmaduke Woodman, Lia Domide,
#   Jochen Mersmann, Anthony R. McIntosh, Viktor Jirsa (2013)
#       The Virtual Brain: a simulator of primate brain network dynamics.
#   Frontiers in Neuroinformatics (7:10. doi: 10.3389/fninf.2013.00010)
#
#

import tempfile

from tvb.interfaces.rest.client.datatype.datatype_api import DataTypeApi
from tvb.interfaces.rest.client.operation.operation_api import OperationApi
from tvb.interfaces.rest.client.project.project_api import ProjectApi
from tvb.interfaces.rest.client.simulator.simulation_api import SimulationApi
from tvb.interfaces.rest.client.user.user_api import UserApi
from tvb.interfaces.rest.commons.dtos import OperationDto


class TVBClient:
    """
    TVB-BrainX3 client class which expose the whole API. Initializing this class with the correct rest server url is mandatory.
    """

    def __init__(self, server_url):
        self.temp_folder = tempfile.gettempdir()
        self.user_api = UserApi(server_url)
        self.project_api = ProjectApi(server_url)
        self.datatype_api = DataTypeApi(server_url)
        self.simulation_api = SimulationApi(server_url)
        self.operation_api = OperationApi(server_url)

    def get_users(self):
        """
        Retrieve users list
        """
        return self.user_api.get_users()

    def get_project_list(self, username):
        """
        Given a username, this function will return all projects for the given user.
        """
        return self.user_api.get_projects_list(username)

    def get_data_in_project(self, project_gid):
        """
        Given a project_gid, this function will return a list of DataTypeDTO instances associated with the current project.
        """
        return self.project_api.get_data_in_project(project_gid)

    def get_operations_in_project(self, project_gid, page_number):
        """
        Given a project_gid and a page number (default page size is 20), this function will return the list of OperationDTO entities
        """
        response = self.project_api.get_operations_in_project(project_gid, page_number)
        operations = [OperationDto(**operation) for operation in response["operations"]]
        pages = response["pages"]
        return operations, pages

    def retrieve_datatype(self, datatype_gid, download_folder):
        """
        Given a guid, this function will download locally the H5 full data from the server to the given folder.
        """

        return self.datatype_api.retrieve_datatype(datatype_gid, download_folder)

    def load_datatype(self, datatype_path):
        """
        TODO: TO BE IMPLEMENTED
        Given a local H5 file location, where previously a valid H5 file has been downloaded from TVB server, load in
        memory a HasTraits subclass instance (e.g. Connectivity, TimeSeriesRegion).
        """
        return self.datatype_api.load_datatype(datatype_path)

    def get_operations_for_datatype(self, datatype_gid):
        """
        Given a guid, this function will return the available operations for that datatype, as a list of AlgorithmDTO instances
        """
        return self.datatype_api.get_operations_for_datatype(datatype_gid)

    def fire_simulation(self, project_gid, session_stored_simulator):
        """
        Given a project to execute the operation, and a configuration for the Simulator’s inputs, this will launch the
        simulation and return its gid
        """
        return self.simulation_api.fire_simulation(project_gid, session_stored_simulator, self.temp_folder)

    def launch_operation(self, project_gid, algorithm_module, algorithm_classname, view_model):
        """
        This is a more generic method of launching Analyzers. Given a project id, algorithm module, algorithm classname
        and a view model instance, this function will serialize the view model and will launch the analyzer.
        """
        return self.operation_api.launch_operation(project_gid, algorithm_module, algorithm_classname,
                                                   view_model, self.temp_folder)

    def get_operation_status(self, operation_gid):
        """
        Given an operation gid, this function returns the status of that operation in TVB. The status of operations can be:
        STATUS_FINISHED = "5-FINISHED"
        STATUS_PENDING = "4-PENDING"
        STATUS_STARTED = "3-STARTED"
        STATUS_CANCELED = "2-CANCELED"
        STATUS_ERROR = "1-ERROR"
        """
        return self.operation_api.get_operation_status(operation_gid)

    def get_operation_results(self, operation_gid):
        """
        Given an operation gid, this function returns a list of DataType instances (subclasses), representing the results of that
        operation if it has finished and an empty list, if the operation is still running, has failed or simply has no results.
        """
        return self.operation_api.get_operations_results(operation_gid)



# -*- coding: utf-8 -*-
#
#
# TheVirtualBrain-Framework Package. This package holds all Data Management, and
# Web-UI helpful to run brain-simulations. To use it, you also need do download
# TheVirtualBrain-Scientific Package (for simulators). See content of the
# documentation-folder for more details. See also http://www.thevirtualbrain.org
#
# (c) 2012-2020, Baycrest Centre for Geriatric Care ("Baycrest") and others
#
# This program is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software Foundation,
# either version 3 of the License, or (at your option) any later version.
# This program is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
# PARTICULAR PURPOSE.  See the GNU General Public License for more details.
# You should have received a copy of the GNU General Public License along with this
# program.  If not, see <http://www.gnu.org/licenses/>.
#
#
#   CITATION:
# When using The Virtual Brain for scientific publications, please cite it as follows:
#
#   Paula Sanz Leon, Stuart A. Knock, M. Marmaduke Woodman, Lia Domide,
#   Jochen Mersmann, Anthony R. McIntosh, Viktor Jirsa (2013)
#       The Virtual Brain: a simulator of primate brain network dynamics.
#   Frontiers in Neuroinformatics (7:10. doi: 10.3389/fninf.2013.00010)
#
#
from tvb.datatypes.cortex import Cortex
from tvb.simulator.coupling import Coupling
from tvb.simulator.integrators import *
from tvb.simulator.models import Model
from tvb.simulator.monitors import Monitor, SubSample, GlobalAverage, TemporalAverage
from tvb.simulator.noise import Additive, Multiplicative, Noise


# TODO: rethink this solution
def config_h5_factory(config_class):
    from tvb.core.entities.file.simulator.cortex_h5 import CortexH5

    if issubclass(config_class, Noise):
        return noise_h5_factory(config_class)
    if issubclass(config_class, Integrator):
        return integrator_h5_factory(config_class)
    if issubclass(config_class, Coupling):
        return coupling_h5_factory(config_class)
    if issubclass(config_class, Model):
        return model_h5_factory(config_class)
    if issubclass(config_class, Monitor):
        return monitor_h5_factory(config_class)
    if issubclass(config_class, Cortex):
        return CortexH5
    return None


def noise_h5_factory(noise_class):
    from tvb.core.entities.file.simulator.noise_h5 import AdditiveH5, MultiplicativeH5

    noise_class_to_h5 = {
        Additive: AdditiveH5,
        Multiplicative: MultiplicativeH5
    }

    return noise_class_to_h5.get(noise_class)


def integrator_h5_factory(integrator_class):
    from tvb.core.entities.file.simulator.integrator_h5 import IntegratorH5, IntegratorStochasticH5

    integrator_class_to_h5 = {
        HeunDeterministic: IntegratorH5,
        HeunStochastic: IntegratorStochasticH5,
        EulerDeterministic: IntegratorH5,
        EulerStochastic: IntegratorStochasticH5,
        RungeKutta4thOrderDeterministic: IntegratorH5,
        Identity: IntegratorH5,
        VODE: IntegratorH5,
        VODEStochastic: IntegratorStochasticH5,
        Dopri5: IntegratorH5,
        Dopri5Stochastic: IntegratorStochasticH5,
        Dop853: IntegratorH5,
        Dop853Stochastic: IntegratorStochasticH5
    }

    return integrator_class_to_h5.get(integrator_class)


def coupling_h5_factory(coupling_class):
    from tvb.simulator.coupling import Linear, Scaling, HyperbolicTangent, Sigmoidal, SigmoidalJansenRit, PreSigmoidal, \
        Difference, Kuramoto
    from tvb.core.entities.file.simulator.coupling_h5 import LinearH5, ScalingH5, HyperbolicTangentH5, SigmoidalH5, \
        SigmoidalJansenRitH5, PreSigmoidalH5, DifferenceH5, KuramotoH5

    coupling_class_to_h5 = {
        Linear: LinearH5,
        Scaling: ScalingH5,
        HyperbolicTangent: HyperbolicTangentH5,
        Sigmoidal: SigmoidalH5,
        SigmoidalJansenRit: SigmoidalJansenRitH5,
        PreSigmoidal: PreSigmoidalH5,
        Difference: DifferenceH5,
        Kuramoto: KuramotoH5
    }

    return coupling_class_to_h5.get(coupling_class)


def model_h5_factory(model_class):
    from tvb.simulator.models import Epileptor, Epileptor2D, EpileptorCodim3, EpileptorCodim3SlowMod, Hopfield, \
        JansenRit, ZetterbergJansen, EpileptorRestingState, LarterBreakspear, Generic2dOscillator, \
        ReducedSetFitzHughNagumo, ReducedSetHindmarshRose, WilsonCowan, ReducedWongWang, ReducedWongWangExcInh, \
        ZerlautFirstOrder, ZerlautSecondOrder, SupHopf, Linear, Kuramoto
    from tvb.core.entities.file.simulator.model_h5 import EpileptorH5, Epileptor2DH5, EpileptorCodim3H5, \
        EpileptorCodim3SlowModH5, HopfieldH5, JansenRitH5, ZetterbergJansenH5, EpileptorRestingStateH5, \
        LarterBreakspearH5, LinearH5, Generic2dOscillatorH5, KuramotoH5, ReducedSetFitzHughNagumoH5, \
        ReducedSetHindmarshRoseH5, WilsonCowanH5, ReducedWongWangH5, ReducedWongWangExcInhH5, ZerlautFirstOrderH5, \
        ZerlautSecondOrderH5, SupHopfH5
    model_class_to_h5 = {
        Epileptor: EpileptorH5,
        Epileptor2D: Epileptor2DH5,
        EpileptorCodim3: EpileptorCodim3H5,
        EpileptorCodim3SlowMod: EpileptorCodim3SlowModH5,
        Hopfield: HopfieldH5,
        JansenRit: JansenRitH5,
        ZetterbergJansen: ZetterbergJansenH5,
        EpileptorRestingState: EpileptorRestingStateH5,
        LarterBreakspear: LarterBreakspearH5,
        Linear: LinearH5,
        Generic2dOscillator: Generic2dOscillatorH5,
        Kuramoto: KuramotoH5,
        ReducedSetFitzHughNagumo: ReducedSetFitzHughNagumoH5,
        ReducedSetHindmarshRose: ReducedSetHindmarshRoseH5,
        WilsonCowan: WilsonCowanH5,
        ReducedWongWang: ReducedWongWangH5,
        ReducedWongWangExcInh: ReducedWongWangExcInhH5,
        ZerlautFirstOrder: ZerlautFirstOrderH5,
        ZerlautSecondOrder: ZerlautSecondOrderH5,
        SupHopf: SupHopfH5
    }

    return model_class_to_h5.get(model_class)


def monitor_h5_factory(model_class):
    from tvb.simulator.monitors import Raw, SpatialAverage, EEG, MEG, iEEG, Bold, BoldRegionROI
    from tvb.core.entities.file.simulator.monitor_h5 import RawH5, SpatialAverageH5, EEGH5, MEGH5, iEEGH5, BoldH5, \
        BoldRegionROIH5, SubSampleH5, GlobalAverageH5, TemporalAverageH5

    monitor_class_to_h5 = {
        Raw: RawH5,
        SubSample: SubSampleH5,
        SpatialAverage: SpatialAverageH5,
        GlobalAverage: GlobalAverageH5,
        TemporalAverage: TemporalAverageH5,
        EEG: EEGH5,
        MEG: MEGH5,
        iEEG: iEEGH5,
        Bold: BoldH5,
        BoldRegionROI: BoldRegionROIH5
    }

    return monitor_class_to_h5.get(model_class)



# -*- coding: utf-8 -*-
#
#
# TheVirtualBrain-Framework Package. This package holds all Data Management, and 
# Web-UI helpful to run brain-simulations. To use it, you also need do download
# TheVirtualBrain-Scientific Package (for simulators). See content of the
# documentation-folder for more details. See also http://www.thevirtualbrain.org
#
# (c) 2012-2020, Baycrest Centre for Geriatric Care ("Baycrest") and others
#
# This program is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software Foundation,
# either version 3 of the License, or (at your option) any later version.
# This program is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
# PARTICULAR PURPOSE.  See the GNU General Public License for more details.
# You should have received a copy of the GNU General Public License along with this
# program.  If not, see <http://www.gnu.org/licenses/>.
#
#
#   CITATION:
# When using The Virtual Brain for scientific publications, please cite it as follows:
#
#   Paula Sanz Leon, Stuart A. Knock, M. Marmaduke Woodman, Lia Domide,
#   Jochen Mersmann, Anthony R. McIntosh, Viktor Jirsa (2013)
#       The Virtual Brain: a simulator of primate brain network dynamics.
#   Frontiers in Neuroinformatics (7:10. doi: 10.3389/fninf.2013.00010)
#
#

"""
.. moduleauthor:: Lia Domide <lia.domide@codemart.ro>
"""

import os
import csv
import sys
import json
import datetime
import uuid
import urllib.request, urllib.parse, urllib.error
from hashlib import md5
import numpy
import six
from tvb.basic.profile import TvbProfile
from tvb.basic.logger.builder import get_logger
from tvb.core.decorators import user_environment_execution

MATLAB = "matlab"
OCTAVE = "octave"

CHAR_SEPARATOR = "__"
CHAR_SPACE = "--"
CHAR_DRIVE = "-DriVe-"
DRIVE_SEP = ":"
COMPLEX_TIME_FORMAT = '%Y-%m-%d,%H-%M-%S.%f'
# LESS_COMPLEX_TIME_FORMAT is also compatible with data exported from TVB 1.0. 
# This is only used as a fallback in the string to date conversion.
LESS_COMPLEX_TIME_FORMAT = '%Y-%m-%d,%H-%M-%S'
SIMPLE_TIME_FORMAT = "%m-%d-%Y"


################## PATH related methods start here ###############

def path2url_part(file_path):
    """
    Prepare a File System Path for passing into an URL.
    """
    if not os.path.isabs(file_path):
        file_path = os.path.join(TvbProfile.current.TVB_STORAGE, file_path)
    result = file_path.replace(os.sep, CHAR_SEPARATOR).replace(" ", CHAR_SPACE).replace(DRIVE_SEP, CHAR_DRIVE)
    return urllib.parse.quote(result)


def url2path(encoded_path):
    """
    Retrieve File System Path from encoded URL (inverse of path2url_part).
    """
    return encoded_path.replace(CHAR_SEPARATOR, os.sep).replace(CHAR_SPACE, " ").replace(CHAR_DRIVE, DRIVE_SEP)


def get_unique_file_name(storage_folder, file_name, try_number=0):
    """
    Compute non-existent file name, in storage_folder.
    Try file_name, and if already exists, try adding a number.
    """
    # TODO this method should be re-tought
    name, ext = os.path.splitext(file_name)
    date = str(datetime.datetime.now())
    date = date.replace(' ', '').replace(':', '').replace('.', '').replace('-', '')
    if try_number > 0:
        file_ = '%s-%s%s' % (name, date, ext)
    else:
        file_ = file_name
    full_path = os.path.join(storage_folder, file_)
    if os.path.exists(full_path):
        # Try another name, by appending the consecutive try_number
        return get_unique_file_name(storage_folder, file_name, try_number + 1)
    return full_path, file_


################## PATH related methods end here ###############    


################## FILE related methods start here ###############

def store_list_data(data_list, file_name, storage_folder, overwrite=False):
    """
    Write a list into a file using CSV writer.
    CSV writer, better than numpy, write also Strings
    """
    if not isinstance(data_list, (list, numpy.ndarray)):
        raise Exception("Invalid given type!! " + str(type(data_list)))
    if overwrite:
        full_path = os.path.join(storage_folder, file_name)
        file_name = os.path.split(full_path)[1]
    else:
        full_path, file_name = get_unique_file_name(storage_folder, file_name)

    # generic writer, capable to write strings also
    with open(full_path, 'wb') as destination:
        csv_writer = csv.writer(destination, delimiter=' ')
        if isinstance(data_list[0], (list, numpy.ndarray)):
            for row in data_list:
                csv_writer.writerow(row)
        else:
            csv_writer.writerow(data_list)

    return file_name


################## FILE related methods end here ###############  


################## CONVERT related methods start here ###############

def parse_json_parameters(parameters):
    """
    From JSON with Unicodes, return a dictionary having strings as keys.
    Loading from DB a JSON will return instead of string keys, unicodes.
    """
    params = json.loads(parameters)
    new_params = {}
    for key, value in six.iteritems(params):
        new_params[str(key)] = value
    return new_params


def string2date(string_input, complex_format=True, date_format=None):
    """Read date from string, after internal format"""
    if date_format is not None:
        return datetime.datetime.strptime(string_input, date_format)
    if complex_format:
        try:
            return datetime.datetime.strptime(string_input, COMPLEX_TIME_FORMAT)
        except ValueError:
            # For backwards compatibility with TVB 1.0
            return datetime.datetime.strptime(string_input, LESS_COMPLEX_TIME_FORMAT)
    return datetime.datetime.strptime(string_input, SIMPLE_TIME_FORMAT)


def date2string(date_input, complex_format=True, date_format=None):
    """Convert date into string, after internal format"""
    if date_input is None:
        return "None"

    if date_format is not None:
        return date_input.strftime(date_format)

    if complex_format:
        return date_input.strftime(COMPLEX_TIME_FORMAT)
    return date_input.strftime(SIMPLE_TIME_FORMAT)


def format_timedelta(timedelta, most_significant2=True):
    """
    Format a datetime.timedelta.
    :param timedelta: object timedelta to format
    :param most_significant2: Will show only the 2 most significant units (ex: hours, minutes). Default True.
    """
    days = timedelta.days
    hours, remainder = divmod(timedelta.seconds, 3600)
    minutes, seconds = divmod(remainder, 60)
    fragments = [str(days) + 'd', str(hours) + 'h', str(minutes) + 'm', str(seconds) + 's']

    if days:
        fragments = fragments[0:]
    elif hours:
        fragments = fragments[1:]
    elif minutes:
        fragments = fragments[2:]
    else:
        fragments = fragments[3:]

    if most_significant2:
        fragments = fragments[:2]

    return ' '.join(fragments)


def string2bool(string_input):
    """ Convert given string into boolean value."""
    string_input = str(string_input).lower()
    return string_input in ("yes", "true", "t", "1")


ARRAY_BEGIN = -1
DATA_UNCONVERTED = 1
DATA_CONVERTED = 2


def string2array(input_data_str, split_char, dtype=None):
    """
    Given an input string first try to load it using JSON and if that fails,
     meaning some weird array is given by the user, fall-back to _custom_string2array.
    """
    prepared_input_data_str = input_data_str.replace(split_char, ',')
    if not prepared_input_data_str.strip().startswith('['):
        # In case of range for a model parameter for example, the input string here is just a float
        # which will be converted to a 0-d numpy array instead of a 1-d array if brackets are missing.
        prepared_input_data_str = '[' + prepared_input_data_str.strip() + ']'
    try:
        array = json.loads(prepared_input_data_str)
        return numpy.array(array)
    except ValueError:
        logger = get_logger(__name__)
        logger.debug("Received input array %s is poorly formatted and could not be evaluated by Python."
                     "Falling back to _custom_string2array." % prepared_input_data_str)
        return _custom_string2array(input_data_str, split_char, dtype)


def _custom_string2array(input_data_str, split_char, dtype=None):
    """
    From a long string, parse a NumPy array.
    """

    class HelperData:
        """Helper for parsing arrays"""

        def __init__(self, data, type_):
            self.data = data
            self.type = type_

    input_str = input_data_str.lstrip().rstrip()
    to_replace = split_char + split_char
    while to_replace in input_str:
        input_str = input_str.replace(to_replace, split_char)
    input_str = input_str.replace('[' + split_char, '[')
    input_str = input_str.replace(split_char + ']', ']')
    input_str = input_str.replace(' ' + split_char, split_char)
    input_str = input_str.replace(split_char + ' ', split_char)
    input_str = input_str.replace('[ ', '[').replace(' ]', ']')

    str_pos = 0
    data_stack = []
    while str_pos < len(input_str):
        if input_str[str_pos] == '[':
            data_stack.append(HelperData('[', -1))
        elif input_str[str_pos] == split_char:
            if data_stack[-1].type == DATA_UNCONVERTED:
                if dtype is not None:
                    elem_type = dtype + "('" + data_stack[-1].data + "')"
                    data_stack[-1].data = eval(elem_type)
                data_stack[-1].type = DATA_CONVERTED
        elif input_str[str_pos] == ']':
            new_array = []
            if data_stack[-1].type == DATA_UNCONVERTED:
                if dtype is not None:
                    elem_type = dtype + "('" + data_stack[-1].data + "')"
                    data_stack[-1].data = eval(elem_type)
                data_stack[-1].type = DATA_CONVERTED
            while data_stack[-1].type != ARRAY_BEGIN:
                last_data = data_stack.pop()
                new_array.insert(0, last_data.data)
            data_stack.pop()
            new_array = HelperData(new_array, 0)
            data_stack.append(new_array)

        elif len(data_stack) == 0 or data_stack[-1].type != DATA_UNCONVERTED:
            data_stack.append(HelperData(input_str[str_pos], DATA_UNCONVERTED))
        else:
            data_stack[-1].data = data_stack[-1].data + input_str[str_pos]
        str_pos += 1
    if len(data_stack) == 0:
        return None

    if type(data_stack[0].data) is str:
        if dtype is not None:
            if data_stack[-1].data == 'None':
                return None
            elem_type = dtype + "('" + data_stack[-1].data + "')"
            data_stack[-1].data = eval(elem_type)
        return data_stack[0].data
    return numpy.array(data_stack[0].data)


class TVBJSONEncoder(json.JSONEncoder):
    """
    Custom encoder class. Referring towards "to_json" method, when found, or default behaviour otherwise.
    """

    def default(self, obj):
        if hasattr(obj, "to_json"):
            return obj.to_json()
        # TODO: Review this quick fix.
        # TVB-2565 numpy int serialization
        if numpy.issubdtype(obj, numpy.integer):
            return int(obj)
        return json.JSONEncoder.default(self, obj)


################## CONVERT related methods end here ###############


################## MATLAB related method start here ###############

def get_matlab_executable():
    """
    Check If MATLAB is installed on current system.
    Return True or False.
    Return True, when MATLAB executable is found in Path.
    """
    matlab_exe_path = None
    if sys.platform.startswith('win'):
        split_char = ";"
        octave_exec = OCTAVE + ".exe"
        matlab_exec = MATLAB + ".exe"
    else:
        split_char = ":"
        octave_exec = OCTAVE
        matlab_exec = MATLAB
    logger = get_logger(__name__)
    logger.debug("Searching Matlab in path: " + str(os.environ["PATH"]))
    for path in os.environ["PATH"].split(split_char):
        if os.path.isfile(os.path.join(path, matlab_exec)):
            matlab_exe_path = os.path.join(path, matlab_exec)
            logger.debug("MATLAB was found:" + path)
            return matlab_exe_path
    for path in os.environ["PATH"].split(split_char):
        if os.path.isfile(os.path.join(path, octave_exec)):
            logger.debug("OCTAVE was found:" + path)
            matlab_exe_path = os.path.join(path, octave_exec)
            return matlab_exe_path
    return matlab_exe_path


@user_environment_execution
def check_matlab_version(matlab_path):
    """
    Try to get the current version of matlab from a given path.
    """
    version = None
    logger = get_logger(__name__)
    matlab_test_file_name = os.path.join(TvbProfile.current.TVB_STORAGE, 'test_mat_version')
    matlab_test_file = matlab_test_file_name + '.m'
    matlab_log_file = os.path.join(TvbProfile.current.TVB_STORAGE, 'version_log.txt')

    try:
        matlab_version_txt = """tvb_checking_version = version
        quit()
        """
        with open(matlab_test_file, 'w') as file_p:
            file_p.write(matlab_version_txt)

        matlab_cmd(matlab_path, matlab_test_file_name, matlab_log_file)

        with open(matlab_log_file) as log_file:
            result_data = log_file.read()
        version = result_data.strip().split('tvb_checking_version')[1]
        version = version.replace('\n', '').strip()

        logger.debug("Response in TVB from: %s\n Version: %s \nOriginal %s" % (matlab_path, version, result_data))
        os.remove(matlab_test_file)
        os.remove(matlab_log_file)

    except Exception:
        logger.exception('Could not parse Matlab Version!')
        try:
            os.remove(matlab_test_file)
            os.remove(matlab_log_file)
        except Exception:
            logger.exception('Could not remove files in the second try...')

    return version


@user_environment_execution
def matlab_cmd(matlab_path, script_name, log_file):
    """
    Called in order to obtain the command for calling MATLAB
    """
    folder_path, exe_name = os.path.split(matlab_path)
    if folder_path not in os.environ['PATH']:
        os.environ['PATH'] = os.environ.get('PATH', '') + os.pathsep + folder_path
    if MATLAB in exe_name:
        base = exe_name + ' '
        opts = ' -nodesktop -nojvm -nosplash -logfile %s -r "run %s "' % (log_file, script_name)
        if os.name == 'nt':
            opts = '-minimize ' + opts
        command = base + opts
    else:  # if OCTAVE in exe_name:
        command = exe_name + ' %s.m >> %s' % (script_name, log_file)
    return os.system(command)


def extract_matlab_doc_string(file_n):
    """
    Extract the first doc entry from a matlab file.
    """
    try:
        with open(file_n) as m_file:
            m_data = m_file.read()
    except Exception:
        return "Description not available."

    doc_started_flag = False
    result = ""

    for row in m_data.split('\n'):
        if row.startswith('%'):
            doc_started_flag = True
            result += row.replace('%', '') + "<br/>"
        else:
            if len(row.strip()) == 0:
                result += "<br/>"
            else:
                if doc_started_flag:
                    break
    return result


################## MATLAB methods end here     ##############

################## GENERIC  methods start here ###############


def generate_guid():
    """ 
    Generate new Global Unique Identifier.
    This identifier should be unique per each station, 
    and unique for different machines.
    """
    return str(uuid.uuid1())


def format_bytes_human(size, si=False):
    """
    :param size: size in kilobytes
    :param si: if True use SI units (multiple of 1000 not 1024)
    :return: a String with [number] [memory unit measure]
    """
    if si:
        m = ['kB', 'MB', 'GB']
        base = 1000.0
    else:
        m = ['KiB', 'MiB', 'GiB']
        base = 1024.0

    exp = 0
    while size >= base and exp < len(m) - 1:
        size /= base
        exp += 1
    return "%.1f %s" % (size, m[exp])


def prepare_time_slice(total_time_length, max_length=10 ** 4):
    """
    Limit the time dimension when retrieving from TS.
    If total time length is greater than MAX, then retrieve only the last part of the TS

    :param total_time_length: TS time dimension
    :param max_length: limiting number of TS steps

    :return: python slice
    """

    if total_time_length < max_length:
        return slice(total_time_length)

    return slice(total_time_length - max_length, total_time_length)


def hash_password(pass_string):
    return md5(pass_string.encode('utf-8')).hexdigest()



# -*- coding: utf-8 -*-
#
#
# TheVirtualBrain-Framework Package. This package holds all Data Management, and 
# Web-UI helpful to run brain-simulations. To use it, you also need do download
# TheVirtualBrain-Scientific Package (for simulators). See content of the
# documentation-folder for more details. See also http://www.thevirtualbrain.org
#
# (c) 2012-2020, Baycrest Centre for Geriatric Care ("Baycrest") and others
#
# This program is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software Foundation,
# either version 3 of the License, or (at your option) any later version.
# This program is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
# PARTICULAR PURPOSE.  See the GNU General Public License for more details.
# You should have received a copy of the GNU General Public License along with this
# program.  If not, see <http://www.gnu.org/licenses/>.
#
#
#   CITATION:
# When using The Virtual Brain for scientific publications, please cite it as follows:
#
#   Paula Sanz Leon, Stuart A. Knock, M. Marmaduke Woodman, Lia Domide,
#   Jochen Mersmann, Anthony R. McIntosh, Viktor Jirsa (2013)
#       The Virtual Brain: a simulator of primate brain network dynamics.
#   Frontiers in Neuroinformatics (7:10. doi: 10.3389/fninf.2013.00010)
#
#

"""
.. moduleauthor:: Lia Domide <lia.domide@codemart.ro>
.. moduleauthor:: Ionel Ortelecan <ionel.ortelecan@codemart.ro>
.. moduleauthor:: Bogdan Neacsa <bogdan.neacsa@codemart.ro>
"""
from tvb.core.adapters.abcadapter import ABCAdapterForm
from tvb.core.entities.model.model_datatype import DataTypeGroup
from tvb.core.entities.model.model_operation import RANGE_MISSING_STRING, STATUS_FINISHED, RANGE_MISSING_VALUE
from tvb.core.entities.storage import dao
from tvb.core.entities.transient.pse import ContextDiscretePSE
from tvb.core.adapters.abcdisplayer import ABCDisplayer
from tvb.adapters.datatypes.db.mapped_value import DatatypeMeasureIndex
from tvb.core.entities.filters.chain import FilterChain
from tvb.core.neotraits.forms import TraitDataTypeSelectField
from tvb.core.neotraits.view_model import ViewModel, DataTypeGidAttr

MAX_NUMBER_OF_POINT_TO_SUPPORT = 512


class DiscretePSEAdapterModel(ViewModel):
    datatype_group = DataTypeGidAttr(
        linked_datatype=DataTypeGroup,
        label='Datatype Group'
    )


class DiscretePSEAdapterForm(ABCAdapterForm):

    def __init__(self, prefix='', project_id=None):
        super(DiscretePSEAdapterForm, self).__init__(prefix, project_id)
        self.datatype_group = TraitDataTypeSelectField(DiscretePSEAdapterModel.datatype_group, self,
                                                       name='datatype_group', conditions=self.get_filters())

    @staticmethod
    def get_view_model():
        return DiscretePSEAdapterModel

    @staticmethod
    def get_required_datatype():
        return DataTypeGroup

    @staticmethod
    def get_input_name():
        return '_datatype_group'

    @staticmethod
    def get_filters():
        return FilterChain(fields=[FilterChain.datatype + ".no_of_ranges", FilterChain.datatype + ".no_of_ranges",
                                   FilterChain.datatype + ".count_results"],
                           operations=["<=", ">=", "<="],
                           values=[2, 1, MAX_NUMBER_OF_POINT_TO_SUPPORT])


class DiscretePSEAdapter(ABCDisplayer):
    """
    Visualization adapter for Parameter Space Exploration.
    Will be used as a generic visualizer, accessible when input entity is DataTypeGroup.
    Will also be used in Burst as a supplementary navigation layer.
    """
    _ui_name = "Discrete Parameter Space Exploration"
    _ui_subsection = "pse"

    def get_form_class(self):
        return DiscretePSEAdapterForm

    def get_required_memory_size(self, view_model):
        # type: (DiscretePSEAdapterModel) -> int
        """
        Return the required memory to run this algorithm.
        """
        # Don't know how much memory is needed.
        return -1

    # TODO: migrate to neotraits
    def launch(self, view_model):
        # type: (DiscretePSEAdapterModel) -> dict
        """
        Launch the visualizer.
        """
        pse_context = self.prepare_parameters(view_model.datatype_group, '')
        pse_context.prepare_individual_jsons()

        return self.build_display_result('pse_discrete/view', pse_context,
                                         pages=dict(controlPage="pse_discrete/controls"))

    @staticmethod
    def prepare_range_labels(operation_group, range_json):
        """
        Prepare Range labels for display in UI.
        When the current range_json is empty, returns None, [RANGE_MISSING_STRING], [RANGE_MISSING_STRING]

        :param operation_group: model.OperationGroup instance
        :param range_json: Stored JSON for for a given range
        :return: String with current range label, Array of ranged numbers, Array of labels for current range
        """
        contains_numbers, range_name, range_values = operation_group.load_range_numbers(range_json)

        if contains_numbers is None:
            return None, range_values, [RANGE_MISSING_STRING], False

        if contains_numbers:
            range_labels = range_values
        else:
            # when datatypes are in range, get the display name for those and use as labels.
            range_labels = []
            for data_gid in range_values:
                range_labels.append(dao.get_datatype_by_gid(data_gid).display_name)

        return range_name, range_values, range_labels, contains_numbers

    @staticmethod
    def get_value_on_axe(op_range, only_numbers, range_param_name, fake_numbers):
        if range_param_name is None:
            return RANGE_MISSING_VALUE
        if only_numbers:
            return op_range[range_param_name]
        return fake_numbers[op_range[range_param_name]]

    @staticmethod
    def prepare_parameters(datatype_group_gid, back_page, color_metric=None, size_metric=None):
        """
        We suppose that there are max 2 ranges and from each operation results exactly one dataType.

        :param datatype_group_gid: the group id for the `DataType` to be visualised
        :param back_page: Page where back button will direct
        :param color_metric: String referring to metric to apply on colors
        :param size_metric:  String referring to metric to apply on sizes

        :returns: `ContextDiscretePSE`
        :raises Exception: when `datatype_group_id` is invalid (not in database)
        """
        datatype_group = dao.get_datatype_group_by_gid(datatype_group_gid.hex)
        if datatype_group is None:
            raise Exception("Selected DataTypeGroup is no longer present in the database. "
                            "It might have been remove or the specified id is not the correct one.")

        operation_group = dao.get_operationgroup_by_id(datatype_group.fk_operation_group)

        name1, values1, labels1, only_numbers1 = DiscretePSEAdapter.prepare_range_labels(operation_group,
                                                                                         operation_group.range1)
        name2, values2, labels2, only_numbers2 = DiscretePSEAdapter.prepare_range_labels(operation_group,
                                                                                         operation_group.range2)

        pse_context = ContextDiscretePSE(datatype_group_gid, color_metric, size_metric, back_page)
        pse_context.setRanges(name1, values1, labels1, name2, values2, labels2,
                              only_numbers1 and only_numbers2)
        final_dict = {}
        operations = dao.get_operations_in_group(operation_group.id)

        fake_numbers1 = dict(list(zip(values1, list(range(len(list(values1)))))))
        fake_numbers2 = dict(list(zip(values2, list(range(len(list(values2)))))))

        for operation_ in operations:
            if not operation_.has_finished:
                pse_context.has_started_ops = True
            range_values = eval(operation_.range_values)
            key_1 = DiscretePSEAdapter.get_value_on_axe(range_values, only_numbers1, name1, fake_numbers1)
            key_2 = DiscretePSEAdapter.get_value_on_axe(range_values, only_numbers2, name2, fake_numbers2)

            datatype = None
            if operation_.status == STATUS_FINISHED:
                pse_filter = FilterChain(fields=[FilterChain.datatype + '.type'], operations=['!='],
                                         values=['SimulatorIndex'])
                datatypes = dao.get_results_for_operation(operation_.id, pse_filter)
                if len(datatypes) > 0:
                    datatype = datatypes[0]
                    if datatype.type == "DatatypeMeasureIndex":
                        # Load proper entity class from DB.
                        measures = dao.get_generic_entity(DatatypeMeasureIndex, datatype.gid)
                    else:
                        measures = dao.get_generic_entity(DatatypeMeasureIndex, datatype.gid, 'source_gid')
                    pse_context.prepare_metrics_datatype(measures, datatype)

            if key_1 not in final_dict:
                final_dict[key_1] = {}

            final_dict[key_1][key_2] = pse_context.build_node_info(operation_, datatype)

        pse_context.fill_object(final_dict)
        # datatypes_dict is not actually used in the drawing of the PSE and actually
        # causes problems in case of NaN values, so just remove it before creating the json
        pse_context.datatypes_dict = {}
        if not only_numbers1:
            pse_context.values_x = list(range(len(list(values1))))
        if not only_numbers2:
            pse_context.values_y = list(range(len(list(values2))))
        return pse_context



# -*- coding: utf-8 -*-
#
#
# TheVirtualBrain-Framework Package. This package holds all Data Management, and
# Web-UI helpful to run brain-simulations. To use it, you also need do download
# TheVirtualBrain-Scientific Package (for simulators). See content of the
# documentation-folder for more details. See also http://www.thevirtualbrain.org
#
# (c) 2012-2020, Baycrest Centre for Geriatric Care ("Baycrest") and others
#
# This program is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software Foundation,
# either version 3 of the License, or (at your option) any later version.
# This program is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
# PARTICULAR PURPOSE.  See the GNU General Public License for more details.
# You should have received a copy of the GNU General Public License along with this
# program.  If not, see <http://www.gnu.org/licenses/>.
#
#
#   CITATION:
# When using The Virtual Brain for scientific publications, please cite it as follows:
#
#   Paula Sanz Leon, Stuart A. Knock, M. Marmaduke Woodman, Lia Domide,
#   Jochen Mersmann, Anthony R. McIntosh, Viktor Jirsa (2013)
#       The Virtual Brain: a simulator of primate brain network dynamics.
#   Frontiers in Neuroinformatics (7:10. doi: 10.3389/fninf.2013.00010)
#
#

"""
Change of DB structure from TVB version 1.3 to 1.3.1

.. moduleauthor:: Mihai Andrei <mihai.andrei@codemart.ro>

"""

from sqlalchemy import Column, Float, Integer
from sqlalchemy.sql import text
from migrate.changeset.schema import create_column, drop_column, alter_column
from tvb.core.entities import model
from tvb.core.entities.storage import SA_SESSIONMAKER
from tvb.basic.logger.builder import get_logger

meta = model.Base.metadata
LOGGER = get_logger(__name__)

COLUMN_N1 = Column('used_disk_space', Float)
COLUMN_N2 = Column('disk_size', Integer)

COLUMN_N3_OLD = Column('result_disk_size', Integer)
COLUMN_N3_NEW = Column('estimated_disk_size', Integer)



def upgrade(migrate_engine):
    """
    Upgrade operations go here.
    Don't create your own engine; bind migrate_engine to your metadata.
    """
    meta.bind = migrate_engine

    table = meta.tables['USERS']
    drop_column(COLUMN_N1, table)
    table = meta.tables['BURST_CONFIGURATIONS']
    drop_column(COLUMN_N2, table)
    table = meta.tables['OPERATIONS']
    alter_column(COLUMN_N3_OLD, table=table, name=COLUMN_N3_NEW.name)

    try:
        meta.bind = migrate_engine
        session = SA_SESSIONMAKER()
        session.execute(text("""UPDATE "DATA_TYPES" SET module='tvb.datatypes.region_mapping' WHERE "type" = 'RegionMapping' """))
        session.execute(text("""UPDATE "DATA_TYPES" SET module='tvb.datatypes.local_connectivity' WHERE "type" = 'LocalConnectivity' """))
        session.execute(text("""UPDATE "DATA_TYPES" SET module='tvb.datatypes.cortex' WHERE "type" = 'Cortex' """))
        session.commit()
        session.close()

    except Exception:
        LOGGER.exception("Cold not update datatypes")
        raise



def downgrade(migrate_engine):
    """Operations to reverse the above upgrade go here."""
    meta.bind = migrate_engine

    table = meta.tables['USERS']
    create_column(COLUMN_N1, table)
    table = meta.tables['BURST_CONFIGURATIONS']
    create_column(COLUMN_N2, table)
    table = meta.tables['OPERATIONS']
    alter_column(COLUMN_N3_NEW, table=table, name=COLUMN_N3_OLD.name)

    try:
        meta.bind = migrate_engine
        session = SA_SESSIONMAKER()
        session.execute(text("""UPDATE "DATA_TYPES" SET module='tvb.datatypes.surfaces' WHERE "type" = 'RegionMapping' """))
        session.execute(text("""UPDATE "DATA_TYPES" SET module='tvb.datatypes.surfaces' WHERE "type" = 'LocalConnectivity' """))
        session.execute(text("""UPDATE "DATA_TYPES" SET module='tvb.datatypes.surfaces' WHERE "type" = 'Cortex' """))
        session.commit()
        session.close()
    except Exception:
        LOGGER.exception("Cold not update datatypes")
        raise




# -*- coding: utf-8 -*-
#
#
# TheVirtualBrain-Framework Package. This package holds all Data Management, and
# Web-UI helpful to run brain-simulations. To use it, you also need do download
# TheVirtualBrain-Scientific Package (for simulators). See content of the
# documentation-folder for more details. See also http://www.thevirtualbrain.org
#
# (c) 2012-2020, Baycrest Centre for Geriatric Care ("Baycrest") and others
#
# This program is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software Foundation,
# either version 3 of the License, or (at your option) any later version.
# This program is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
# PARTICULAR PURPOSE.  See the GNU General Public License for more details.
# You should have received a copy of the GNU General Public License along with this
# program.  If not, see <http://www.gnu.org/licenses/>.
#
#
#   CITATION:
# When using The Virtual Brain for scientific publications, please cite it as follows:
#
#   Paula Sanz Leon, Stuart A. Knock, M. Marmaduke Woodman, Lia Domide,
#   Jochen Mersmann, Anthony R. McIntosh, Viktor Jirsa (2013)
#       The Virtual Brain: a simulator of primate brain network dynamics.
#   Frontiers in Neuroinformatics (7:10. doi: 10.3389/fninf.2013.00010)
#
#

"""
Backend-side for TS Visualizer of TS Volume DataTypes.

.. moduleauthor:: Robert Parcus <betoparcus@gmail.com>
.. moduleauthor:: Lia Domide <lia.domide@codemart.ro>
.. moduleauthor:: Ciprian Tomoiaga <ciprian.tomoiaga@codemart.ro>

"""

import json
import numpy
from tvb.adapters.visualizers.region_volume_mapping import _MappedArrayVolumeBase
from tvb.core.adapters.abcadapter import ABCAdapterForm
from tvb.core.adapters.arguments_serialisation import preprocess_space_parameters, postprocess_voxel_ts
from tvb.core.entities.filters.chain import FilterChain
from tvb.adapters.datatypes.h5.time_series_h5 import TimeSeriesVolumeH5, TimeSeriesRegionH5
from tvb.adapters.datatypes.db.structural import StructuralMRIIndex
from tvb.adapters.datatypes.db.time_series import TimeSeriesIndex
from tvb.core.entities.storage import dao
from tvb.core.neotraits.forms import TraitDataTypeSelectField
from tvb.core.neotraits.view_model import ViewModel, DataTypeGidAttr
from tvb.core.utils import prepare_time_slice
from tvb.datatypes.structural import StructuralMRI
from tvb.datatypes.time_series import TimeSeries


class TimeSeriesVolumeVisualiserModel(ViewModel):
    time_series = DataTypeGidAttr(
        linked_datatype=TimeSeries,
        label='Time Series'
    )

    background = DataTypeGidAttr(
        linked_datatype=StructuralMRI,
        required=False,
        label='Background T1'
    )


class TimeSeriesVolumeVisualiserForm(ABCAdapterForm):

    def __init__(self, prefix='', project_id=None):
        super(TimeSeriesVolumeVisualiserForm, self).__init__(prefix, project_id)
        self.time_series = TraitDataTypeSelectField(TimeSeriesVolumeVisualiserModel.time_series, self,
                                                    name='time_series', conditions=self.get_filters())
        self.background = TraitDataTypeSelectField(TimeSeriesVolumeVisualiserModel.background, self, name='background')

    @staticmethod
    def get_view_model():
        return TimeSeriesVolumeVisualiserModel

    @staticmethod
    def get_input_name():
        return '_time_series'

    @staticmethod
    def get_filters():
        return FilterChain(fields=[FilterChain.datatype + '.has_volume_mapping'], operations=["=="], values=[True])

    @staticmethod
    def get_required_datatype():
        return TimeSeriesIndex


class TimeSeriesVolumeVisualiser(_MappedArrayVolumeBase):
    _ui_name = "Time Series Volume Visualizer"
    _ui_subsection = "volume"

    def get_form_class(self):
        return TimeSeriesVolumeVisualiserForm

    def get_required_memory_size(self, view_model):
        # type: (TimeSeriesVolumeVisualiserModel) -> int
        """Return required memory."""
        return -1

    def launch(self, view_model):
        # type: (TimeSeriesVolumeVisualiserModel) -> dict

        url_volume_data = self.build_url('get_volume_view', view_model.time_series.hex, '')
        url_timeseries_data = self.build_url('get_voxel_time_series', view_model.time_series.hex, '')

        ts_h5_class, ts_h5_path = self._load_h5_of_gid(view_model.time_series.hex)
        ts_h5 = ts_h5_class(ts_h5_path)
        min_value, max_value = ts_h5.get_min_max_values()

        if isinstance(ts_h5, TimeSeriesVolumeH5):
            volume_h5_class, volume_h5_path = self._load_h5_of_gid(ts_h5.volume.load().hex)
            volume_h5 = volume_h5_class(volume_h5_path)
            volume_shape = ts_h5.data.shape
        else:
            rmv_index = self.load_entity_by_gid(ts_h5.region_mapping_volume.load(True).hex)
            rmv_h5_class, rmv_h5_path = self._load_h5_of_gid(rmv_index.gid)
            rmv_h5 = rmv_h5_class(rmv_h5_path)
            volume_index = self.load_entity_by_gid(rmv_h5.volume.load().hex)
            volume_h5_class, volume_h5_path = self._load_h5_of_gid(volume_index.gid)
            volume_h5 = volume_h5_class(volume_h5_path)
            volume_shape = [ts_h5.data.shape[0]]
            volume_shape.extend(rmv_h5.array_data.shape)
            rmv_h5.close()

        background_index = None
        if view_model.background:
            background_index = self.load_entity_by_gid(view_model.background.hex)

        params = dict(title="Volumetric Time Series",
                      ts_title=ts_h5.title.load(),
                      labelsStateVar=ts_h5.labels_dimensions.load().get(ts_h5.labels_ordering.load()[1], []),
                      labelsModes=list(range(ts_h5.data.shape[3])),
                      minValue=min_value, maxValue=max_value,
                      urlVolumeData=url_volume_data,
                      urlTimeSeriesData=url_timeseries_data,
                      samplePeriod=ts_h5.sample_period.load(),
                      samplePeriodUnit=ts_h5.sample_period_unit.load(),
                      volumeShape=json.dumps(volume_shape),
                      volumeOrigin=json.dumps(volume_h5.origin.load().tolist()),
                      voxelUnit=volume_h5.voxel_unit.load(),
                      voxelSize=json.dumps(volume_h5.voxel_size.load().tolist()))

        params.update(self.ensure_background(background_index))

        volume_h5.close()
        ts_h5.close()
        return self.build_display_result("time_series_volume/view", params,
                                         pages=dict(controlPage="time_series_volume/controls"))

    def ensure_background(self, background_index):
        if background_index is None:
            background_index = dao.try_load_last_entity_of_type(self.current_project_id, StructuralMRIIndex)

        if background_index is None:
            return _MappedArrayVolumeBase.compute_background_params()

        background_class, background_path = self._load_h5_of_gid(background_index.gid)
        background_h5 = background_class(background_path)
        min_value, max_value = background_h5.get_min_max_values()
        background_h5.close()

        url_volume_data = self.build_url('get_volume_view', background_index.gid, '')
        return _MappedArrayVolumeBase.compute_background_params(min_value, max_value, url_volume_data)

    def get_voxel_time_series(self, entity_gid, **kwargs):
        """
        Retrieve for a given voxel (x,y,z) the entire timeline.

        :param x: int coordinate
        :param y: int coordinate
        :param z: int coordinate

        :return: A complex dictionary with information about current voxel.
                The main part will be a vector with all the values over time from the x,y,z coordinates.
        """

        ts_h5_class, ts_h5_path = self._load_h5_of_gid(entity_gid)

        with ts_h5_class(ts_h5_path) as ts_h5:
            if ts_h5_class is TimeSeriesRegionH5:
                return self._get_voxel_time_series_region(ts_h5, **kwargs)

            return ts_h5.get_voxel_time_series(**kwargs)

    def _get_voxel_time_series_region(self, ts_h5, x, y, z, var=0, mode=0):
        region_mapping_volume_gid = ts_h5.region_mapping_volume.load()
        if region_mapping_volume_gid is None:
            raise Exception("Invalid method called for TS without Volume Mapping!")

        volume_rm_h5_class, volume_rm_h5_path = self._load_h5_of_gid(region_mapping_volume_gid.hex)
        volume_rm_h5 = volume_rm_h5_class(volume_rm_h5_path)

        volume_rm_shape = volume_rm_h5.array_data.shape
        x, y, z = preprocess_space_parameters(x, y, z, volume_rm_shape[0], volume_rm_shape[1], volume_rm_shape[2])
        idx_slices = slice(x, x + 1), slice(y, y + 1), slice(z, z + 1)

        idx = int(volume_rm_h5.array_data[idx_slices])

        time_length = ts_h5.data.shape[0]
        var, mode = int(var), int(mode)
        voxel_slices = prepare_time_slice(time_length), slice(var, var + 1), slice(idx, idx + 1), slice(mode, mode + 1)

        connectivity_gid = volume_rm_h5.connectivity.load()
        connectivity_h5_class, connectivity_h5_path = self._load_h5_of_gid(connectivity_gid.hex)
        connectivity_h5 = connectivity_h5_class(connectivity_h5_path)
        label = connectivity_h5.region_labels.load()[idx]

        background, back_min, back_max = None, None, None
        if idx < 0:
            back_min, back_max = ts_h5.get_min_max_values()
            background = numpy.ones((time_length, 1)) * ts_h5.out_of_range(back_min)
            label = 'background'

        volume_rm_h5.close()
        connectivity_h5.close()

        result = postprocess_voxel_ts(self, voxel_slices, background, back_min, back_max, label)
        return result



# -*- coding: utf-8 -*-
#
#
# TheVirtualBrain-Framework Package. This package holds all Data Management, and
# Web-UI helpful to run brain-simulations. To use it, you also need do download
# TheVirtualBrain-Scientific Package (for simulators). See content of the
# documentation-folder for more details. See also http://www.thevirtualbrain.org
#
# (c) 2012-2020, Baycrest Centre for Geriatric Care ("Baycrest") and others
#
# This program is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software Foundation,
# either version 3 of the License, or (at your option) any later version.
# This program is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
# PARTICULAR PURPOSE.  See the GNU General Public License for more details.
# You should have received a copy of the GNU General Public License along with this
# program.  If not, see <http://www.gnu.org/licenses/>.
#
#
#   CITATION:
# When using The Virtual Brain for scientific publications, please cite it as follows:
#
#   Paula Sanz Leon, Stuart A. Knock, M. Marmaduke Woodman, Lia Domide,
#   Jochen Mersmann, Anthony R. McIntosh, Viktor Jirsa (2013)
#       The Virtual Brain: a simulator of primate brain network dynamics.
#   Frontiers in Neuroinformatics (7:10. doi: 10.3389/fninf.2013.00010)
#
#

from tvb.simulator.monitors import Monitor, Raw, SpatialAverage, Projection, EEG, MEG, iEEG, Bold
from tvb.core.entities.file.simulator.configurations_h5 import SimulatorConfigurationH5
from tvb.core.neotraits.h5 import Scalar, DataSet, Reference, EquationScalar


class MonitorH5(SimulatorConfigurationH5):

    def __init__(self, path):
        super(MonitorH5, self).__init__(path)
        self.period = Scalar(Monitor.period, self)
        self.variables_of_interest = DataSet(Monitor.variables_of_interest, self)


class RawH5(MonitorH5):

    def __init__(self, path):
        super(RawH5, self).__init__(path)
        self.period = Scalar(Raw.period, self)
        self.variables_of_interest = DataSet(Raw.variables_of_interest, self)


class SubSampleH5(MonitorH5):
    """"""


class SpatialAverageH5(MonitorH5):

    def __init__(self, path):
        super(SpatialAverageH5, self).__init__(path)
        self.spatial_mask = DataSet(SpatialAverage.spatial_mask, self)
        self.default_mask = Scalar(SpatialAverage.default_mask, self)


class GlobalAverageH5(MonitorH5):
    """"""


class TemporalAverageH5(MonitorH5):
    """"""


class ProjectionH5(MonitorH5):

    def __init__(self, path):
        super(ProjectionH5, self).__init__(path)
        self.region_mapping = Reference(Projection.region_mapping, self)
        self.obnoise = Reference(Projection.obsnoise, self)

    def store(self, datatype, scalars_only=False, store_references=False):
        # type: (Projection, bool, bool) -> None
        super(ProjectionH5, self).store(datatype, scalars_only, store_references)
        self.region_mapping.store(datatype.region_mapping.gid)


class EEGH5(ProjectionH5):

    def __init__(self, path):
        super(EEGH5, self).__init__(path)
        self.projection = Reference(EEG.projection, self)
        self.sensors = Reference(EEG.sensors, self)
        self.reference = Scalar(EEG.reference, self)
        self.sigma = Scalar(EEG.sigma, self)

    def store(self, datatype, scalars_only=False, store_references=False):
        # type: (Projection, bool, bool) -> None
        super(EEGH5, self).store(datatype, scalars_only, store_references)
        self.projection.store(datatype.projection.gid)
        self.sensors.store(datatype.sensors.gid)


class MEGH5(ProjectionH5):

    def __init__(self, path):
        super(MEGH5, self).__init__(path)
        self.projection = Reference(MEG.projection, self)
        self.sensors = Reference(MEG.sensors, self)

    def store(self, datatype, scalars_only=False, store_references=False):
        # type: (Projection, bool, bool) -> None
        super(MEGH5, self).store(datatype, scalars_only, store_references)
        self.projection.store(datatype.projection.gid)
        self.sensors.store(datatype.sensors.gid)


class iEEGH5(ProjectionH5):

    def __init__(self, path):
        super(iEEGH5, self).__init__(path)
        self.projection = Reference(iEEG.projection, self)
        self.sensors = Reference(iEEG.sensors, self)
        self.sigma = Scalar(iEEG.sigma, self)

    def store(self, datatype, scalars_only=False, store_references=False):
        # type: (Projection, bool, bool) -> None
        super(iEEGH5, self).store(datatype, scalars_only, store_references)
        self.projection.store(datatype.projection.gid)
        self.sensors.store(datatype.sensors.gid)


class BoldH5(MonitorH5):

    def __init__(self, path):
        super(BoldH5, self).__init__(path)
        self.period = Scalar(Bold.period, self)
        self.hrf_kernel = EquationScalar(Bold.hrf_kernel, self)
        self.hrf_length = Scalar(Bold.hrf_length, self)


class BoldRegionROIH5(BoldH5):
    """"""



# -*- coding: utf-8 -*-
#
#
# TheVirtualBrain-Framework Package. This package holds all Data Management, and 
# Web-UI helpful to run brain-simulations. To use it, you also need do download
# TheVirtualBrain-Scientific Package (for simulators). See content of the
# documentation-folder for more details. See also http://www.thevirtualbrain.org
#
# (c) 2012-2020, Baycrest Centre for Geriatric Care ("Baycrest") and others
#
# This program is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software Foundation,
# either version 3 of the License, or (at your option) any later version.
# This program is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
# PARTICULAR PURPOSE.  See the GNU General Public License for more details.
# You should have received a copy of the GNU General Public License along with this
# program.  If not, see <http://www.gnu.org/licenses/>.
#
#
#   CITATION:
# When using The Virtual Brain for scientific publications, please cite it as follows:
#
#   Paula Sanz Leon, Stuart A. Knock, M. Marmaduke Woodman, Lia Domide,
#   Jochen Mersmann, Anthony R. McIntosh, Viktor Jirsa (2013)
#       The Virtual Brain: a simulator of primate brain network dynamics.
#   Frontiers in Neuroinformatics (7:10. doi: 10.3389/fninf.2013.00010)
#
#

"""
.. Ionel Ortelecan <ionel.ortelecan@codemart.ro>
"""

import uuid
from tvb.adapters.datatypes.db.connectivity import ConnectivityIndex
from tvb.adapters.datatypes.db.surface import SurfaceIndex
from tvb.adapters.simulator.equation_forms import get_form_for_equation
from tvb.core.adapters.abcadapter import ABCSynchronous, ABCAdapterForm
from tvb.core.neocom import h5
from tvb.core.neotraits.forms import DataTypeSelectField, FormField, SimpleStrField, TraitDataTypeSelectField, \
    SelectField
from tvb.core.neotraits.view_model import ViewModel, DataTypeGidAttr
from tvb.datatypes.connectivity import Connectivity
from tvb.datatypes.equations import Sigmoid, PulseTrain
from tvb.datatypes.patterns import StimuliSurface, StimuliRegion
from tvb.adapters.datatypes.db.patterns import StimuliRegionIndex, StimuliSurfaceIndex
from tvb.datatypes.surfaces import CorticalSurface
from tvb.interfaces.web.controllers.decorators import using_template


class StimulusSurfaceSelectorForm(ABCAdapterForm):

    def __init__(self, project_id):
        super(StimulusSurfaceSelectorForm, self).__init__()
        self.project_id = project_id
        self.surface_stimulus = DataTypeSelectField(StimuliSurfaceIndex, self, name='existentEntitiesSelect',
                                                    label='Load Surface Stimulus')
        self.display_name = SimpleStrField(self, name='display_name', label='Display name')

    @using_template('spatial/spatial_fragment')
    def __str__(self):
        return {'form': self, 'legend': 'Loaded stimulus'}


class SurfaceStimulusCreatorModel(ViewModel, StimuliSurface):
    surface = DataTypeGidAttr(
        linked_datatype=CorticalSurface,
        label=StimuliSurface.surface.label
    )


class SurfaceStimulusCreatorForm(ABCAdapterForm):
    NAME_SPATIAL_PARAMS_DIV = 'spatial_params'
    NAME_TEMPORAL_PARAMS_DIV = 'temporal_params'
    default_spatial = Sigmoid
    default_temporal = PulseTrain

    def __init__(self, spatial_equation_choices, temporal_equation_choices, project_id):
        super(SurfaceStimulusCreatorForm, self).__init__()
        self.project_id = project_id

        # TODO: filter CorticalSurafces
        self.surface = TraitDataTypeSelectField(SurfaceStimulusCreatorModel.surface, self, name='surface',
                                                conditions=self.get_filters())
        self.spatial = SelectField(SurfaceStimulusCreatorModel.spatial, self, name='spatial',
                                   choices=spatial_equation_choices)
        self.spatial_params = FormField(get_form_for_equation(self.default_spatial), self,
                                        name=self.NAME_SPATIAL_PARAMS_DIV)
        self.temporal = SelectField(SurfaceStimulusCreatorModel.temporal, self, name='temporal',
                                    choices=temporal_equation_choices)
        self.temporal_params = FormField(get_form_for_equation(self.default_temporal), self,
                                         name=self.NAME_TEMPORAL_PARAMS_DIV)

    @staticmethod
    def get_view_model():
        return SurfaceStimulusCreatorModel

    @staticmethod
    def get_required_datatype():
        return SurfaceIndex

    @staticmethod
    def get_input_name():
        return '_surface'

    @staticmethod
    def get_filters():
        return None

    def fill_from_trait(self, trait):
        self.surface.data = trait.surface.hex
        self.spatial.data = type(trait.spatial)
        self.temporal.data = type(trait.temporal)
        self.spatial_params.form = get_form_for_equation(type(trait.spatial))(self.NAME_SPATIAL_PARAMS_DIV)
        self.temporal_params.form = get_form_for_equation(type(trait.temporal))(self.NAME_TEMPORAL_PARAMS_DIV)
        self.spatial_params.form.fill_from_trait(trait.spatial)
        self.temporal_params.form.fill_from_trait(trait.temporal)

    @using_template('spatial/spatial_fragment')
    def __str__(self):
        return {'form': self, 'next_action': 'form_spatial_surface_stimulus_equations',
                'spatial_params_div': self.NAME_SPATIAL_PARAMS_DIV,
                'temporal_params_div': self.NAME_TEMPORAL_PARAMS_DIV, 'legend': 'Stimulus interface'}


class SurfaceStimulusCreator(ABCSynchronous):
    """
    The purpose of this adapter is to create a StimuliSurface.
    """
    KEY_SURFACE = 'surface'
    KEY_SPATIAL = 'spatial'
    KEY_TEMPORAL = 'temporal'
    KEY_FOCAL_POINTS_TRIANGLES = 'focal_points_triangles'

    def get_form_class(self):
        return SurfaceStimulusCreatorForm

    def get_output(self):
        """
        Describes the outputs of the launch method.
        """
        return [StimuliSurfaceIndex]

    @staticmethod
    def prepare_stimuli_surface_from_view_model(view_model, load_full_surface=False):
        # type: (SurfaceStimulusCreatorModel, bool) -> StimuliSurface
        stimuli_surface = StimuliSurface()

        stimuli_surface.focal_points_triangles = view_model.focal_points_triangles
        stimuli_surface.spatial = view_model.spatial
        stimuli_surface.temporal = view_model.temporal

        surface_index = SurfaceStimulusCreator.load_entity_by_gid(view_model.surface.hex)
        if load_full_surface:
            stimuli_surface.surface = h5.load_from_index(surface_index, CorticalSurface)
        else:
            stimuli_surface.surface = CorticalSurface()
            stimuli_surface.gid = view_model.surface
            surface_h5 = h5.h5_file_for_index(surface_index)
            # We need to load surface triangles on stimuli because focal_points_surface property needs to acces them
            stimuli_surface.surface.triangles = surface_h5.triangles.load()
            surface_h5.close()

        return stimuli_surface

    def launch(self, view_model):
        # type: (SurfaceStimulusCreatorModel) -> [StimuliSurfaceIndex]
        """
        Used for creating a `StimuliSurface` instance
        """
        stimuli_surface = self.prepare_stimuli_surface_from_view_model(view_model, view_model.surface)
        stimuli_surface_index = StimuliSurfaceIndex()
        stimuli_surface_index.fill_from_has_traits(stimuli_surface)

        h5.store_complete(stimuli_surface, self.storage_path)
        return stimuli_surface_index

    def get_required_memory_size(self, view_model):
        # type: (SurfaceStimulusCreatorModel) -> int
        """
        Return the required memory to run this algorithm.
        """
        return -1

    def get_required_disk_size(self, view_model):
        # type: (SurfaceStimulusCreatorModel) -> int
        """
        Returns the required disk size to be able to run the adapter. (in kB)
        """
        return 0


class StimulusRegionSelectorForm(ABCAdapterForm):

    def __init__(self, project_id):
        super(StimulusRegionSelectorForm, self).__init__()
        self.project_id = project_id
        self.region_stimulus = DataTypeSelectField(StimuliRegionIndex, self, name='existentEntitiesSelect',
                                                   label='Load Region Stimulus')
        self.display_name = SimpleStrField(self, name='display_name', label='Display name')

    @using_template('spatial/spatial_fragment')
    def __str__(self):
        return {'form': self, 'legend': 'Loaded stimulus'}


class RegionStimulusCreatorModel(ViewModel, StimuliRegion):
    connectivity = DataTypeGidAttr(
        field_type=uuid.UUID,
        linked_datatype=Connectivity,
        label="Connectivity"
    )


class RegionStimulusCreatorForm(ABCAdapterForm):
    NAME_TEMPORAL_PARAMS_DIV = 'temporal_params'
    default_temporal = PulseTrain

    def __init__(self, equation_choices, project_id):
        super(RegionStimulusCreatorForm, self).__init__()
        self.project_id = project_id

        self.connectivity = TraitDataTypeSelectField(RegionStimulusCreatorModel.connectivity, self, name='connectivity')
        self.temporal = SelectField(RegionStimulusCreatorModel.temporal, self, name='temporal', choices=equation_choices)
        self.temporal_params = FormField(get_form_for_equation(self.default_temporal), self,
                                         name=self.NAME_TEMPORAL_PARAMS_DIV)
        # self.temporal.template = 'form_fields/select_field.html'

    @staticmethod
    def get_view_model():
        return RegionStimulusCreatorModel

    @staticmethod
    def get_filters():
        return None

    @staticmethod
    def get_input_name():
        return '_connectivity'

    @staticmethod
    def get_required_datatype():
        return ConnectivityIndex

    def fill_from_trait(self, trait):
        # type: (RegionStimulusCreatorModel) -> None
        self.connectivity.data = trait.connectivity.hex
        self.temporal.data = type(trait.temporal)
        self.temporal_params.form = get_form_for_equation(type(trait.temporal))(self.NAME_TEMPORAL_PARAMS_DIV)
        self.temporal_params.form.fill_from_trait(trait.temporal)

    @using_template('spatial/spatial_fragment')
    def __str__(self):
        return {'form': self, 'next_action': 'form_spatial_model_param_equations',
                'temporal_params_div': self.NAME_TEMPORAL_PARAMS_DIV, 'legend': 'Stimulus interface'}


class RegionStimulusCreator(ABCSynchronous):
    """
    The purpose of this adapter is to create a StimuliRegion.
    """
    def get_form_class(self):
        return RegionStimulusCreatorForm

    def get_output(self):
        """
        Describes the outputs of the launch method.
        """
        return [StimuliRegionIndex]

    def launch(self, view_model):
        # type: (RegionStimulusCreatorModel) -> [StimuliRegionIndex]
        """
        Used for creating a `StimuliRegion` instance
        """
        stimuli_region = StimuliRegion()
        stimuli_region.connectivity = Connectivity()
        stimuli_region.connectivity.gid = view_model.connectivity
        stimuli_region.weight = view_model.weight
        stimuli_region.temporal = view_model.temporal

        stimuli_region_idx = StimuliRegionIndex()
        stimuli_region_idx.fill_from_has_traits(stimuli_region)

        h5.store_complete(stimuli_region, self.storage_path)
        return stimuli_region_idx

    def get_required_disk_size(self, view_model):
        # type: (RegionStimulusCreatorModel) -> int
        """
        Returns the required disk size to be able to run the adapter. (in kB)
        """
        return 0

    def get_required_memory_size(self, view_model):
        # type: (RegionStimulusCreatorModel) -> int
        """
        Return the required memory to run this algorithm.
        """
        return -1



# -*- coding: utf-8 -*-
#
#
# TheVirtualBrain-Framework Package. This package holds all Data Management, and
# Web-UI helpful to run brain-simulations. To use it, you also need do download
# TheVirtualBrain-Scientific Package (for simulators). See content of the
# documentation-folder for more details. See also http://www.thevirtualbrain.org
#
# (c) 2012-2020, Baycrest Centre for Geriatric Care ("Baycrest") and others
#
# This program is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software Foundation,
# either version 3 of the License, or (at your option) any later version.
# This program is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
# PARTICULAR PURPOSE.  See the GNU General Public License for more details.
# You should have received a copy of the GNU General Public License along with this
# program.  If not, see <http://www.gnu.org/licenses/>.
#
#
#   CITATION:
# When using The Virtual Brain for scientific publications, please cite it as follows:
#
#   Paula Sanz Leon, Stuart A. Knock, M. Marmaduke Woodman, Lia Domide,
#   Jochen Mersmann, Anthony R. McIntosh, Viktor Jirsa (2013)
#       The Virtual Brain: a simulator of primate brain network dynamics.
#   Frontiers in Neuroinformatics (7:10. doi: 10.3389/fninf.2013.00010)
#
#

"""
Change of DB structure from TVB version 1.0.6 to 1.0.8

.. moduleauthor:: Lia Domide <lia.domide@codemart.ro>
"""

from sqlalchemy.sql import text
from tvb.basic.logger.builder import get_logger
from tvb.core.entities.storage import SA_SESSIONMAKER


def upgrade(_migrate_engine):
    """
    Upgrade operations go here.
    Don't create your own engine; bind migrate_engine to your metadata.
    """
    try:
        session = SA_SESSIONMAKER()
        session.execute(text("""UPDATE "BURST_CONFIGURATIONS" SET _simulator_configuration =
                                REPLACE(REPLACE(_simulator_configuration, "first_range", "range_1"),
                                                                          "second_range", "range_2");"""))
        session.execute(text("""UPDATE "OPERATIONS" SET parameters =
                                REPLACE(REPLACE(parameters, "first_range", "range_1"), "second_range", "range_2");"""))
        session.commit()
        session.close()
    except Exception as excep:
        ## This update is not critical. We can run even in case of error at update
        logger = get_logger(__name__)
        logger.exception(excep)


def downgrade(_migrate_engine):
    """Operations to reverse the above upgrade go here."""
    try:
        session = SA_SESSIONMAKER()
        session.execute(text("""UPDATE "BURST_CONFIGURATIONS" SET _simulator_configuration =
                                REPLACE(REPLACE(_simulator_configuration, "range_1", "first_range"),
                                                                          "range_2", "second_range");"""))
        session.execute(text("""UPDATE "OPERATIONS" SET parameters =
                                REPLACE(REPLACE(parameters, "range_1", "first_range"), "range_2", "second_range");"""))
        session.commit()
        session.close()
    except Exception as excep:
        ## This update is not critical. We can run even in case of error at update
        logger = get_logger(__name__)
        logger.exception(excep)





from enum import Enum


class Strings(Enum):
    PAGE_NUMBER = "page"
    BASE_PATH = "api"


class RestNamespace(Enum):
    USERS = "/users"
    PROJECTS = "/projects"
    DATATYPES = "/datatypes"
    OPERATIONS = "/operations"
    SIMULATION = "/simulation"


class LinkPlaceholder(Enum):
    USERNAME = "username"
    PROJECT_GID = "project_gid"
    DATATYPE_GID = "datatype_gid"
    OPERATION_GID = "operation_gid"
    ALG_MODULE = "algorithm_module"
    ALG_CLASSNAME = "algorithm_classname"


class RestLink(Enum):
    # USERS
    PROJECTS_LIST = "/{" + LinkPlaceholder.USERNAME.value + "}/projects"

    # PROJECTS
    DATA_IN_PROJECT = "/{" + LinkPlaceholder.PROJECT_GID.value + "}/data"
    OPERATIONS_IN_PROJECT = "/{" + LinkPlaceholder.PROJECT_GID.value + "}/operations"

    # DATATYPES
    GET_DATATYPE = "/{" + LinkPlaceholder.DATATYPE_GID.value + "}"
    DATATYPE_OPERATIONS = "/{" + LinkPlaceholder.DATATYPE_GID.value + "}/operations"

    # OPERATIONS
    LAUNCH_OPERATION = "/{" + LinkPlaceholder.PROJECT_GID.value + "}/algorithm/{" + LinkPlaceholder.ALG_MODULE.value + "}/{" + LinkPlaceholder.ALG_CLASSNAME.value + "}"
    OPERATION_STATUS = "/{" + LinkPlaceholder.OPERATION_GID.value + "}/status"
    OPERATION_RESULTS = "/{" + LinkPlaceholder.OPERATION_GID.value + "}/results"

    # SIMULATION
    FIRE_SIMULATION = "/{" + LinkPlaceholder.PROJECT_GID.value + "}"

    def compute_url(self, include_namespace=False, values=None):
        if values is None:
            values = {}
        result = ""
        if include_namespace:
            for namespace, links in _namespace_url_dict.items():
                if self in links:
                    result += namespace.value
                    break
        result += self.value
        return result.format(**values)


_namespace_url_dict = {
    RestNamespace.USERS: [RestLink.PROJECTS_LIST],
    RestNamespace.PROJECTS: [RestLink.DATA_IN_PROJECT, RestLink.OPERATIONS_IN_PROJECT],
    RestNamespace.DATATYPES: [RestLink.GET_DATATYPE, RestLink.DATATYPE_OPERATIONS],
    RestNamespace.OPERATIONS: [RestLink.LAUNCH_OPERATION, RestLink.OPERATION_STATUS, RestLink.OPERATION_RESULTS],
    RestNamespace.SIMULATION: [RestLink.FIRE_SIMULATION]
}



# -*- coding: utf-8 -*-
#
#
# TheVirtualBrain-Framework Package. This package holds all Data Management, and 
# Web-UI helpful to run brain-simulations. To use it, you also need do download
# TheVirtualBrain-Scientific Package (for simulators). See content of the
# documentation-folder for more details. See also http://www.thevirtualbrain.org
#
# (c) 2012-2020, Baycrest Centre for Geriatric Care ("Baycrest") and others
#
# This program is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software Foundation,
# either version 3 of the License, or (at your option) any later version.
# This program is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
# PARTICULAR PURPOSE.  See the GNU General Public License for more details.
# You should have received a copy of the GNU General Public License along with this
# program.  If not, see <http://www.gnu.org/licenses/>.
#
#
#   CITATION:
# When using The Virtual Brain for scientific publications, please cite it as follows:
#
#   Paula Sanz Leon, Stuart A. Knock, M. Marmaduke Woodman, Lia Domide,
#   Jochen Mersmann, Anthony R. McIntosh, Viktor Jirsa (2013)
#       The Virtual Brain: a simulator of primate brain network dynamics.
#   Frontiers in Neuroinformatics (7:10. doi: 10.3389/fninf.2013.00010)
#
#
"""
Change of DB structure from TVB version 1.0.1 to TVB 1.0.2.

.. moduleauthor:: Bogdan Neacsa <bogdan.neacsa@codemart.ro>
"""

import sqlalchemy
from sqlalchemy import Column, Integer, String
from sqlalchemy.sql import text
from migrate.changeset.schema import create_column, drop_column
from tvb.core.entities import model
from tvb.core.entities.storage import SA_SESSIONMAKER

meta = model.Base.metadata
COL_1 = Column('_length_1d', Integer)
COL_2 = Column('_length_2d', Integer)
COL_3 = Column('_length_3d', Integer)
COL_4 = Column('_length_4d', Integer)
COL_5 = Column('_labels_dimensions', String, default="{}")
COL_6 = Column('_labels_ordering', String, default="[]")
COL_7 = Column('_unidirectional', Integer)

TABLE_RENAMES = [('BURST_CONFIGURATION', 'BURST_CONFIGURATIONS'),
                 ('CONNECTIVITY_SELECTION', 'CONNECTIVITY_SELECTIONS'),
                 ('USER_TO_PROJECT', 'USERS_TO_PROJECTS'),
                 ('OPERATION_PROCESS_IDENTIFIER', 'OPERATION_PROCESS_IDENTIFIERS')]



def upgrade(migrate_engine):
    """
    Upgrade operations go here.
    Don't create your own engine; bind migrate_engine to your metadata.
    """
    meta.bind = migrate_engine
    table1 = meta.tables['MAPPED_TIME_SERIES_DATA']

    create_column(COL_1, table1)
    create_column(COL_2, table1)
    create_column(COL_3, table1)
    create_column(COL_4, table1)
    create_column(COL_5, table1)

    session = SA_SESSIONMAKER()
    try:
        # We have a database that supports renaming columns. This way we save data from old timeseries.
        session.execute(text("ALTER TABLE \"MAPPED_TIME_SERIES_DATA\" "
                             "RENAME COLUMN _dim_labels to _labels_ordering"))
        session.execute(text("ALTER TABLE \"MAPPED_CROSS_CORRELATION_DATA\" "
                             "RENAME COLUMN _dim_labels to _labels_ordering"))
    except sqlalchemy.exc.OperationalError:
        # We have a database like sqlite. Just create a new column, we're gonna miss old data in this case.
        session.execute(text("ALTER TABLE \"MAPPED_TIME_SERIES_DATA\" "
                             "ADD COLUMN _labels_ordering VARYING CHARACTER(255)"))
        session.execute(text("ALTER TABLE \"MAPPED_CROSS_CORRELATION_DATA\" "
                             "ADD COLUMN _labels_ordering VARYING CHARACTER(255)"))

    session.execute(text("DROP TABLE \"MAPPED_PSI_TABLE_DATA\""))
    session.execute(text("DROP TABLE \"MAPPED_NERF_TABLE_DATA\""))
    session.execute(text("DROP TABLE \"MAPPED_LOOK_UP_TABLES_DATA\""))
    session.commit()
    session.close()

    table2 = meta.tables['MAPPED_CONNECTIVITY_DATA']
    create_column(COL_7, table2)

    for mapping in TABLE_RENAMES:
        session = SA_SESSIONMAKER()
        session.execute(text("ALTER TABLE \"%s\" RENAME TO \"%s\"" % (mapping[0], mapping[1])))
        session.commit()
        session.close()



def downgrade(migrate_engine):
    """Operations to reverse the above upgrade go here."""
    meta.bind = migrate_engine

    table1 = meta.tables['MAPPED_TIME_SERIES_DATA']
    drop_column(COL_1, table1)
    drop_column(COL_2, table1)
    drop_column(COL_3, table1)
    drop_column(COL_4, table1)

    session = SA_SESSIONMAKER()
    try:
        session.execute(text("ALTER TABLE \"MAPPED_TIME_SERIES_DATA\" "
                             "RENAME COLUMN _labels_ordering to _dim_labels"))
        session.execute(text("ALTER TABLE \"MAPPED_CROSS_CORRELATION_DATA\" "
                             "RENAME COLUMN _labels_ordering to _dim_labels"))

    except sqlalchemy.exc.OperationalError:
        session.execute(text("ALTER TABLE \"MAPPED_TIME_SERIES_DATA\" "
                             "ADD COLUMN _dim_labels VARYING CHARACTER(255)"))
        session.execute(text("ALTER TABLE \"MAPPED_CROSS_CORRELATION_DATA\" "
                             "ADD COLUMN _dim_labels VARYING CHARACTER(255)"))
    session.commit()
    session.close()

    table2 = meta.tables['MAPPED_CONNECTIVITY_DATA']
    drop_column(COL_7, table2)

    for mapping in TABLE_RENAMES:
        session = SA_SESSIONMAKER()
        session.execute(text("ALTER TABLE \"%s\" RENAME TO \"%s\"" % (mapping[1], mapping[0])))
        session.commit()
        session.close()
        
        
        


# -*- coding: utf-8 -*-
#
#
# TheVirtualBrain-Framework Package. This package holds all Data Management, and 
# Web-UI helpful to run brain-simulations. To use it, you also need do download
# TheVirtualBrain-Scientific Package (for simulators). See content of the
# documentation-folder for more details. See also http://www.thevirtualbrain.org
#
# (c) 2012-2020, Baycrest Centre for Geriatric Care ("Baycrest") and others
#
# This program is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software Foundation,
# either version 3 of the License, or (at your option) any later version.
# This program is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
# PARTICULAR PURPOSE.  See the GNU General Public License for more details.
# You should have received a copy of the GNU General Public License along with this
# program.  If not, see <http://www.gnu.org/licenses/>.
#
#
#   CITATION:
# When using The Virtual Brain for scientific publications, please cite it as follows:
#
#   Paula Sanz Leon, Stuart A. Knock, M. Marmaduke Woodman, Lia Domide,
#   Jochen Mersmann, Anthony R. McIntosh, Viktor Jirsa (2013)
#       The Virtual Brain: a simulator of primate brain network dynamics.
#   Frontiers in Neuroinformatics (7:10. doi: 10.3389/fninf.2013.00010)
#
#

"""
.. moduleauthor:: Lia Domide <lia.domide@codemart.ro>
.. moduleauthor:: Bogdan Neacsa <bogdan.neacsa@codemart.ro>
.. moduleauthor:: Yann Gordon <yann@invalid.tvb>
"""

import os
import sys
import signal
import queue as queue
import threading
from subprocess import Popen, PIPE
from tvb.basic.profile import TvbProfile
from tvb.basic.logger.builder import get_logger
from tvb.core.services.burst_service import BurstService
from tvb.core.utils import parse_json_parameters
from tvb.core.entities.model.model_operation import OperationProcessIdentifier, STATUS_ERROR, STATUS_CANCELED
from tvb.core.entities.storage import dao


LOGGER = get_logger(__name__)

CURRENT_ACTIVE_THREADS = []

LOCKS_QUEUE = queue.Queue(0)
for i in range(TvbProfile.current.MAX_THREADS_NUMBER):
    LOCKS_QUEUE.put(1)


class OperationExecutor(threading.Thread):
    """
    Thread in charge for starting an operation, used both on cluster and with stand-alone installations.
    """


    def __init__(self, op_id):
        threading.Thread.__init__(self)
        self.operation_id = op_id
        self._stop_ev = threading.Event()


    def run(self):
        """
        Get the required data from the operation queue and launch the operation.
        """
        # Try to get a spot to launch own operation.
        LOCKS_QUEUE.get(True)
        operation_id = self.operation_id
        run_params = [TvbProfile.current.PYTHON_INTERPRETER_PATH, '-m', 'tvb.core.operation_async_launcher',
                      str(operation_id), TvbProfile.CURRENT_PROFILE_NAME]

        # In the exceptional case where the user pressed stop while the Thread startup is done,
        # We should no longer launch the operation.
        if self.stopped() is False:

            env = os.environ.copy()
            env['PYTHONPATH'] = os.pathsep.join(sys.path)
            # anything that was already in $PYTHONPATH should have been reproduced in sys.path

            launched_process = Popen(run_params, stdout=PIPE, stderr=PIPE, env=env)

            LOGGER.debug("Storing pid=%s for operation id=%s launched on local machine." % (operation_id,
                                                                                            launched_process.pid))
            op_ident = OperationProcessIdentifier(operation_id, pid=launched_process.pid)
            dao.store_entity(op_ident)

            if self.stopped():
                # In the exceptional case where the user pressed stop while the Thread startup is done.
                # and stop_operation is concurrently asking about OperationProcessIdentity.
                self.stop_pid(launched_process.pid)

            subprocess_result = launched_process.communicate()
            LOGGER.info("Finished with launch of operation %s" % operation_id)
            returned = launched_process.wait()

            if returned != 0 and not self.stopped():
                # Process did not end as expected. (e.g. Segmentation fault)
                burst_service = BurstService()
                operation = dao.get_operation_by_id(self.operation_id)
                LOGGER.error("Operation suffered fatal failure! Exit code: %s Exit message: %s" % (returned,
                                                                                                   subprocess_result))

                burst_service.persist_operation_state(operation, STATUS_ERROR,
                                                         "Operation failed unexpectedly! Please check the log files.")

                burst_entity = dao.get_burst_for_operation_id(self.operation_id)
                if burst_entity:
                    message = "Error in operation process! Possibly segmentation fault."
                    burst_service.mark_burst_finished(burst_entity, error_message=message)

            del launched_process

        # Give back empty spot now that you finished your operation
        CURRENT_ACTIVE_THREADS.remove(self)
        LOCKS_QUEUE.put(1)


    def _stop(self):
        """ Mark current thread for stop"""
        self._stop_ev.set()


    def stopped(self):
        """Check if current thread was marked for stop."""
        return self._stop_ev.isSet()


    @staticmethod
    def stop_pid(pid):
        """
        Stop a process specified by PID.
        :returns: True when specified Process was stopped in here, \
                  False in case of exception(e.g. process stopped in advance).
        """
        if sys.platform == 'win32':
            try:
                import ctypes

                handle = ctypes.windll.kernel32.OpenProcess(1, False, int(pid))
                ctypes.windll.kernel32.TerminateProcess(handle, -1)
                ctypes.windll.kernel32.CloseHandle(handle)
            except OSError:
                return False
        else:
            try:
                os.kill(int(pid), signal.SIGKILL)
            except OSError:
                return False

        return True


class StandAloneClient(object):
    """
    Instead of communicating with a back-end cluster, fire locally a new thread.
    """


    @staticmethod
    def execute(operation_id, user_name_label, adapter_instance):
        """Start asynchronous operation locally"""
        thread = OperationExecutor(operation_id)
        CURRENT_ACTIVE_THREADS.append(thread)
        thread.start()


    @staticmethod
    def stop_operation(operation_id):
        """
        Stop a thread for a given operation id
        """
        operation = dao.try_get_operation_by_id(operation_id)
        if not operation or operation.has_finished:
            LOGGER.warning("Operation already stopped or not found is given to stop job: %s" % operation_id)
            return True

        LOGGER.debug("Stopping operation: %s" % str(operation_id))

        # Set the thread stop flag to true
        for thread in CURRENT_ACTIVE_THREADS:
            if int(thread.operation_id) == operation_id:
                thread._stop()
                LOGGER.debug("Found running thread for operation: %d" % operation_id)

        # Kill Thread
        stopped = True
        operation_process = dao.get_operation_process_for_operation(operation_id)
        if operation_process is not None:
            # Now try to kill the operation if it exists
            stopped = OperationExecutor.stop_pid(operation_process.pid)
            if not stopped:
                LOGGER.debug("Operation %d was probably killed from it's specific thread." % operation_id)
            else:
                LOGGER.debug("Stopped OperationExecutor process for %d" % operation_id)

        # Mark operation as canceled in DB and on disk
        BurstService().persist_operation_state(operation, STATUS_CANCELED)

        return stopped


class ClusterSchedulerClient(object):
    # TODO: fix this with neoforms
    """
    Simple class, to mimic the same behavior we are expecting from StandAloneClient, but firing behind
    the cluster job scheduling process..
    """


    @staticmethod
    def _run_cluster_job(operation_identifier, user_name_label, adapter_instance):
        """
        Threaded Popen
        It is the function called by the ClusterSchedulerClient in a Thread.
        This function starts a new process.
        """
        # Load operation so we can estimate the execution time
        operation = dao.get_operation_by_id(operation_identifier)
        kwargs = parse_json_parameters(operation.parameters)
        # kwargs = adapter_instance.prepare_ui_inputs(kwargs)
        time_estimate = int(adapter_instance.get_execution_time_approximation(**kwargs))
        hours = int(time_estimate / 3600)
        minutes = (int(time_estimate) % 3600) / 60
        seconds = int(time_estimate) % 60
        # Anything lower than 5 hours just use default walltime
        if hours < 5:
            walltime = "05:00:00"
        else:
            if hours < 10:
                hours = "0%d" % hours
            else:
                hours = str(hours)
            walltime = "%s:%s:%s" % (hours, str(minutes), str(seconds))

        call_arg = TvbProfile.current.cluster.SCHEDULE_COMMAND % (operation_identifier, user_name_label, walltime)
        LOGGER.info(call_arg)
        process_ = Popen([call_arg], stdout=PIPE, shell=True)
        job_id = process_.stdout.read().replace('\n', '').split(TvbProfile.current.cluster.JOB_ID_STRING)[-1]
        LOGGER.info("Got jobIdentifier = %s for CLUSTER operationID = %s" % (job_id, operation_identifier))
        operation_identifier = OperationProcessIdentifier(operation_identifier, job_id=job_id)
        dao.store_entity(operation_identifier)


    @staticmethod
    def execute(operation_id, user_name_label, adapter_instance):
        """Call the correct system command to submit a job to the cluster."""
        thread = threading.Thread(target=ClusterSchedulerClient._run_cluster_job,
                                  kwargs={'operation_identifier': operation_id,
                                          'user_name_label': user_name_label,
                                          'adapter_instance': adapter_instance})
        thread.start()


    @staticmethod
    def stop_operation(operation_id):
        """
        Stop a thread for a given operation id
        """
        operation = dao.try_get_operation_by_id(operation_id)
        if not operation or operation.has_finished:
            LOGGER.warning("Operation already stopped or not found is given to stop job: %s" % operation_id)
            return True

        operation_process = dao.get_operation_process_for_operation(operation_id)
        result = 0
        # Try to kill only if operation job process is not None
        if operation_process is not None:
            stop_command = TvbProfile.current.cluster.STOP_COMMAND % operation_process.job_id
            LOGGER.info("Stopping cluster operation: %s" % stop_command)
            result = os.system(stop_command)
            if result != 0:
                LOGGER.error("Stopping cluster operation was unsuccessful. Try following status with '" +
                             TvbProfile.current.cluster.STATUS_COMMAND + "'" % operation_process.job_id)

        BurstService().persist_operation_state(operation, STATUS_CANCELED)

        return result == 0


if TvbProfile.current.cluster.IS_DEPLOY:
    # Return an entity capable to submit jobs to the cluster.
    BACKEND_CLIENT = ClusterSchedulerClient()
else:
    # Return a thread launcher.
    BACKEND_CLIENT = StandAloneClient()



# -*- coding: utf-8 -*-
#
#
# TheVirtualBrain-Framework Package. This package holds all Data Management, and 
# Web-UI helpful to run brain-simulations. To use it, you also need do download
# TheVirtualBrain-Scientific Package (for simulators). See content of the
# documentation-folder for more details. See also http://www.thevirtualbrain.org
#
# (c) 2012-2020, Baycrest Centre for Geriatric Care ("Baycrest") and others
#
# This program is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software Foundation,
# either version 3 of the License, or (at your option) any later version.
# This program is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
# PARTICULAR PURPOSE.  See the GNU General Public License for more details.
# You should have received a copy of the GNU General Public License along with this
# program.  If not, see <http://www.gnu.org/licenses/>.
#
#
#   CITATION:
# When using The Virtual Brain for scientific publications, please cite it as follows:
#
#   Paula Sanz Leon, Stuart A. Knock, M. Marmaduke Woodman, Lia Domide,
#   Jochen Mersmann, Anthony R. McIntosh, Viktor Jirsa (2013)
#       The Virtual Brain: a simulator of primate brain network dynamics.
#   Frontiers in Neuroinformatics (7:10. doi: 10.3389/fninf.2013.00010)
#
#

"""
Adapter that uses the traits module to generate interfaces for FFT Analyzer.

.. moduleauthor:: Lia Domide <lia.domide@codemart.ro>
.. moduleauthor:: Stuart A. Knock <Stuart@tvb.invalid>

"""
import uuid
import psutil
import numpy
import math
import tvb.analyzers.fft as fft
import tvb.core.adapters.abcadapter as abcadapter
from tvb.core.entities.filters.chain import FilterChain
from tvb.core.neotraits.view_model import ViewModel, DataTypeGidAttr
from tvb.datatypes.time_series import TimeSeries
from tvb.adapters.datatypes.h5.spectral_h5 import FourierSpectrumH5
from tvb.adapters.datatypes.db.spectral import FourierSpectrumIndex
from tvb.adapters.datatypes.db.time_series import TimeSeriesIndex
from tvb.core.neotraits.forms import ScalarField, TraitDataTypeSelectField, SelectField
from tvb.core.neocom import h5


class FFTAdapterModel(ViewModel, fft.FFT):
    """
    Parameters have the following meaning:
    - time_series: the input time series to which the fft is to be applied
    - segment_length: the block size which determines the frequency resolution of the resulting power spectra
    - window_function: windowing functions can be applied before the FFT is performed
    - detrend: None; specify if detrend is performed on the time series
    """
    time_series = DataTypeGidAttr(
        linked_datatype=TimeSeries,
        label="Time Series",
        doc="""The TimeSeries to which the FFT is to be applied."""
    )


class FFTAdapterForm(abcadapter.ABCAdapterForm):

    def __init__(self, prefix='', project_id=None):
        super(FFTAdapterForm, self).__init__(prefix, project_id)
        self.time_series = TraitDataTypeSelectField(FFTAdapterModel.time_series, self, name='time_series',
                                                    conditions=self.get_filters(), has_all_option=True)
        self.segment_length = ScalarField(FFTAdapterModel.segment_length, self)
        self.window_function = SelectField(FFTAdapterModel.window_function, self)
        self.detrend = ScalarField(FFTAdapterModel.detrend, self)

    @staticmethod
    def get_view_model():
        return FFTAdapterModel

    @staticmethod
    def get_required_datatype():
        return TimeSeriesIndex

    @staticmethod
    def get_filters():
        return FilterChain(fields=[FilterChain.datatype + '.data_ndim'], operations=["=="], values=[4])

    @staticmethod
    def get_input_name():
        return "time_series"

    def get_traited_datatype(self):
        return fft.FFT()


class FourierAdapter(abcadapter.ABCAsynchronous):
    """ TVB adapter for calling the FFT algorithm. """

    _ui_name = "Fourier Spectral Analysis"
    _ui_description = "Calculate the FFT of a TimeSeries entity."
    _ui_subsection = "fourier"

    def __init__(self):
        super(FourierAdapter, self).__init__()
        self.algorithm = fft.FFT()
        self.memory_factor = 1

    def get_form_class(self):
        return FFTAdapterForm

    def get_output(self):
        return [FourierSpectrumIndex]

    def configure(self, view_model):
        # type: (FFTAdapterModel) -> None
        """
        Do any configuration needed before launching.
        """
        self.input_time_series_index = self.load_entity_by_gid(view_model.time_series.hex)
        self.input_shape = (self.input_time_series_index.data_length_1d,
                            self.input_time_series_index.data_length_2d,
                            self.input_time_series_index.data_length_3d,
                            self.input_time_series_index.data_length_4d)

        self.log.debug("time_series shape is %s" % str(self.input_shape))
        self.log.debug("Provided segment_length is %s" % view_model.segment_length)
        self.log.debug("Provided window_function is %s" % view_model.window_function)
        self.log.debug("Detrend is %s" % view_model.detrend)
        # -------------------- Fill Algorithm for Analysis -------------------
        # The enumerate set function isn't working well. A get around strategy is to create a new algorithm
        if view_model.segment_length is not None:
            self.algorithm.segment_length = view_model.segment_length

        self.algorithm.window_function = view_model.window_function
        self.algorithm.detrend = view_model.detrend

        self.log.debug("Using segment_length is %s" % self.algorithm.segment_length)
        self.log.debug("Using window_function  is %s" % self.algorithm.window_function)
        self.log.debug("Using detrend  is %s" % self.algorithm.detrend)

    def get_required_memory_size(self, view_model):
        # type: (FFTAdapterModel) -> int
        """
        Returns the required memory to be able to run the adapter.
        """
        input_size = numpy.prod(self.input_shape) * 8.0
        output_size = self.algorithm.result_size(self.input_shape, self.algorithm.segment_length,
                                                 self.input_time_series_index.sample_period)
        total_free_memory = psutil.virtual_memory().free + psutil.swap_memory().free
        total_required_memory = input_size + output_size
        while total_required_memory / self.memory_factor / total_free_memory > 0.8:
            self.memory_factor += 1
        return total_required_memory / self.memory_factor

    def get_required_disk_size(self, view_model):
        # type: (FFTAdapterModel) -> int
        """
        Returns the required disk size to be able to run the adapter (in kB).
        """
        output_size = self.algorithm.result_size(self.input_shape, self.algorithm.segment_length,
                                                 self.input_time_series_index.sample_period)
        return self.array_size2kb(output_size)

    def launch(self, view_model):
        # type: (FFTAdapterModel) -> [FourierSpectrumIndex]
        """
        Launch algorithm and build results.

        :param time_series: the input time series to which the fft is to be applied
        :param segment_length: the block size which determines the frequency resolution \
                               of the resulting power spectra
        :param window_function: windowing functions can be applied before the FFT is performed
        :type  window_function: None; ‘hamming’; ‘bartlett’; ‘blackman’; ‘hanning’
        :returns: the fourier spectrum for the specified time series
        :rtype: `FourierSpectrumIndex`

        """
        fft_index = FourierSpectrumIndex()
        fft_index.source_gid = view_model.time_series.hex

        block_size = int(math.floor(self.input_shape[2] / self.memory_factor))
        blocks = int(math.ceil(self.input_shape[2] / block_size))

        input_time_series_h5 = h5.h5_file_for_index(self.input_time_series_index)

        dest_path = h5.path_for(self.storage_path, FourierSpectrumH5, fft_index.gid)
        spectra_file = FourierSpectrumH5(dest_path)
        spectra_file.gid.store(uuid.UUID(fft_index.gid))
        spectra_file.source.store(uuid.UUID(self.input_time_series_index.gid))

        # ------------- NOTE: Assumes 4D, Simulator timeSeries. --------------
        node_slice = [slice(self.input_shape[0]), slice(self.input_shape[1]), None, slice(self.input_shape[3])]

        # ---------- Iterate over slices and compose final result ------------
        small_ts = TimeSeries()
        small_ts.sample_period = input_time_series_h5.sample_period.load()

        for block in range(blocks):
            node_slice[2] = slice(block * block_size, min([(block + 1) * block_size, self.input_shape[2]]), 1)
            small_ts.data = input_time_series_h5.read_data_slice(tuple(node_slice))
            self.algorithm.time_series = small_ts
            partial_result = self.algorithm.evaluate()

            if blocks <= 1 and len(partial_result.array_data) == 0:
                self.add_operation_additional_info(
                    "Fourier produced empty result (most probably due to a very short input TimeSeries).")
                return None
            spectra_file.write_data_slice(partial_result)
        fft_index.ndim = len(spectra_file.array_data.shape)
        input_time_series_h5.close()

        fft_index.windowing_function = self.algorithm.window_function
        fft_index.segment_length = self.algorithm.segment_length
        fft_index.detrend = self.algorithm.detrend
        fft_index.frequency_step = partial_result.freq_step
        fft_index.max_frequency = partial_result.max_freq

        spectra_file.segment_length.store(self.algorithm.segment_length)
        spectra_file.windowing_function.store(str(self.algorithm.window_function))
        spectra_file.close()

        self.log.debug("partial segment_length is %s" % (str(partial_result.segment_length)))
        return fft_index



# -*- coding: utf-8 -*-
#
#
# TheVirtualBrain-Framework Package. This package holds all Data Management, and 
# Web-UI helpful to run brain-simulations. To use it, you also need do download
# TheVirtualBrain-Scientific Package (for simulators). See content of the
# documentation-folder for more details. See also http://www.thevirtualbrain.org
#
# (c) 2012-2020, Baycrest Centre for Geriatric Care ("Baycrest") and others
#
# This program is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software Foundation,
# either version 3 of the License, or (at your option) any later version.
# This program is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
# PARTICULAR PURPOSE.  See the GNU General Public License for more details.
# You should have received a copy of the GNU General Public License along with this
# program.  If not, see <http://www.gnu.org/licenses/>.
#
#
#   CITATION:
# When using The Virtual Brain for scientific publications, please cite it as follows:
#
#   Paula Sanz Leon, Stuart A. Knock, M. Marmaduke Woodman, Lia Domide,
#   Jochen Mersmann, Anthony R. McIntosh, Viktor Jirsa (2013)
#       The Virtual Brain: a simulator of primate brain network dynamics.
#   Frontiers in Neuroinformatics (7:10. doi: 10.3389/fninf.2013.00010)
#
#

"""
.. moduleauthor:: Mihai Andrei <mihai.andrei@codemart.ro>
"""

import uuid
from tvb.basic.neotraits.api import Attr
from tvb.core.adapters.abcuploader import ABCUploader, ABCUploaderForm
from tvb.basic.logger.builder import get_logger
from tvb.core.adapters.exceptions import ParseException, LaunchException
from tvb.adapters.datatypes.h5.graph_h5 import ConnectivityMeasureH5
from tvb.adapters.datatypes.db.graph import ConnectivityMeasureIndex
from tvb.core.entities.storage import transactional
from tvb.core.neotraits.forms import TraitUploadField, StrField, TraitDataTypeSelectField
from tvb.core.neotraits.db import from_ndarray
from tvb.core.neocom import h5
from tvb.core.neotraits.uploader_view_model import UploaderViewModel
from tvb.core.neotraits.view_model import Str, DataTypeGidAttr
from tvb.datatypes.connectivity import Connectivity


class ConnectivityMeasureImporterModel(UploaderViewModel):
    data_file = Str(
        label='Connectivity measure file (.mat format)'
    )

    dataset_name = Attr(
        field_type=str,
        default='M',
        label='Matlab dataset name',
        doc='Name of the MATLAB dataset where data is stored'
    )

    connectivity = DataTypeGidAttr(
        linked_datatype=Connectivity,
        label='Large Scale Connectivity',
        doc='The Connectivity for which these measurements were made'
    )


class ConnectivityMeasureImporterForm(ABCUploaderForm):

    def __init__(self, prefix='', project_id=None):
        super(ConnectivityMeasureImporterForm, self).__init__(prefix, project_id)

        self.data_file = TraitUploadField(ConnectivityMeasureImporterModel.data_file, '.mat', self, name='data_file')
        self.dataset_name = StrField(ConnectivityMeasureImporterModel.dataset_name, self, name='dataset_name')
        self.connectivity = TraitDataTypeSelectField(ConnectivityMeasureImporterModel.connectivity, self,
                                                     name='connectivity')

    @staticmethod
    def get_view_model():
        return ConnectivityMeasureImporterModel


class ConnectivityMeasureImporter(ABCUploader):
    """
    This imports a searies of conectivity measures from a .mat file
    """
    _ui_name = "ConnectivityMeasure"
    _ui_subsection = "connectivity_measure"
    _ui_description = "Import a searies of connectivity measures from a .mat file"

    def get_form_class(self):
        return ConnectivityMeasureImporterForm

    def get_output(self):
        return [ConnectivityMeasureIndex]

    @transactional
    def launch(self, view_model):
        # type: (ConnectivityMeasureImporterModel) -> [ConnectivityMeasureIndex]
        """
        Execute import operations:
        """
        try:
            data = self.read_matlab_data(view_model.data_file, view_model.dataset_name)
            measurement_count, node_count = data.shape
            connectivity_index = self.load_entity_by_gid(view_model.connectivity.hex)

            if node_count != connectivity_index.number_of_regions:
                raise LaunchException('The measurements are for %s nodes but the selected connectivity'
                                      ' contains %s nodes' % (node_count, connectivity_index.number_of_regions))

            measures = []
            for i in range(measurement_count):
                cm_idx = ConnectivityMeasureIndex()
                cm_idx.type = ConnectivityMeasureIndex.__name__
                cm_idx.connectivity_gid = connectivity_index.gid

                cm_data = data[i, :]
                cm_idx.array_data_ndim = cm_data.ndim
                cm_idx.ndim = cm_data.ndim
                cm_idx.array_data_min, cm_idx.array_data_max, cm_idx.array_data_mean = from_ndarray(cm_data)

                cm_h5_path = h5.path_for(self.storage_path, ConnectivityMeasureH5, cm_idx.gid)
                with ConnectivityMeasureH5(cm_h5_path) as cm_h5:
                    cm_h5.array_data.store(data[i, :])
                    cm_h5.connectivity.store(uuid.UUID(connectivity_index.gid))
                    cm_h5.gid.store(uuid.UUID(cm_idx.gid))

                cm_idx.user_tag_2 = "nr.-%d" % (i + 1)
                cm_idx.user_tag_3 = "conn_%d" % node_count
                measures.append(cm_idx)
            return measures
        except ParseException as excep:
            logger = get_logger(__name__)
            logger.exception(excep)
            raise LaunchException(excep)
